import logging
import os
import re
import pathlib
import time

import h5py
import numpy as np
from qtpy import QtCore
from PIL import Image

from xypattern.auto_background import SmoothBrucknerBackground
from xypattern import Pattern

logger = logging.getLogger(__name__)


class BatchModel(QtCore.QObject):
    """
    Class describe a model for batch integration
    """

    def __init__(self, configuration):
        super(BatchModel, self).__init__()

        self.data = None
        self.bkg = None
        self.binning = None
        self.file_map = None
        self.files = None
        self.pos_map = None
        self.pos_map_all = None
        self.n_img = None
        self.n_img_all = None
        self.raw_available = False

        self.configuration = configuration
        self.used_mask = None
        self.used_mask_shape = None
        self.used_calibration = None

    def reset_data(self):
        self.data = None
        self.bkg = None
        self.binning = None
        self.file_map = None
        self.files = None
        self.pos_map = None
        self.pos_map_all = None
        self.n_img = None
        self.n_img_all = None
        self.used_mask = None
        self.used_mask_shape = None
        self.used_calibration = None
        self.raw_available = False

    def set_image_files(self, files):
        """
        Set internal variables with respect of given list of files.

        Open each file and count number of images inside. Position of each image in the file
        and total number of images are stored in internal variables.

        :param files: List of file names including path
        """
        if files is None:
            return
        pos_map = []
        file_map = [0]
        image_counter = 0

        self.configuration.img_model.blockSignals(True)

        for i, file in enumerate(files):
            # Assume tif file contains only one image
            if file[-4:] == ".tif":
                n_img = 1
            else:
                if not os.path.exists(file):
                    return
                self.configuration.img_model.load(file)
                n_img = self.configuration.img_model.series_max
            image_counter += n_img
            pos_map += list(zip([i] * n_img, range(n_img)))
            file_map.append(image_counter)

        self.configuration.img_model.blockSignals(False)

        self.files = np.array(files)
        self.n_img_all = image_counter
        self.raw_available = True
        self.pos_map_all = np.array(pos_map)
        self.file_map = np.array(file_map)

    def try_load_old_format(self, data_file):
        self.data = data_file["data"][()]
        self.binning = data_file["binning"][()]
        self.file_map = data_file["file_map"][()]
        self.files = data_file["files"][()].astype("U")
        self.pos_map = data_file["pos_map"][()]
        self.n_img = self.data.shape[0]
        self.n_img_all = self.data.shape[0]
        logger.info("Loading data using deprecated format")

        try:
            cal_file = str(data_file.attrs["calibration"])
            if os.path.isfile(cal_file):
                self.calibration_model.load(cal_file)
        except KeyError:
            logger.info("Calibration info is not found")

        if "mask" in data_file.attrs:
            try:
                mask_file = data_file.attrs["mask"]
                self.mask_model.load_mask(mask_file)
            except FileNotFoundError:
                logger.info("Mask file is not found")

        if "bkg" in data_file:
            self.data = data_file["bkg"][()]

    def load_proc_data(self, filename):
        """
        Load diffraction patterns and metadata from h5 file

        """
        with h5py.File(filename, "r") as data_file:
            # ToDo To be removed
            if "processed/result" not in data_file:
                self.try_load_old_format(data_file)
                return
            self.data = data_file["processed/result/data"][()]
            self.binning = data_file["processed/result/binning"][()]
            self.n_img = self.data.shape[0]
            self.n_img_all = self.data.shape[0]

            if "process" not in data_file["processed"]:
                logger.info("No matching to raw data")
                return

            self.file_map = data_file["processed/process/file_map"][()]
            self.files = data_file["processed/process/files"][()].astype("U")
            self.pos_map = data_file["processed/process/pos_map"][()]

            if isinstance(data_file["processed/process/cal_file"][()], bytes):
                self.used_calibration = str(
                    data_file["processed/process/cal_file"][()].decode("utf-8")
                )
            else:
                self.used_calibration = str(data_file["processed/process/cal_file"][()])
            if os.path.isfile(self.used_calibration):
                self.configuration.calibration_model.load(self.used_calibration)

            if "mask" in data_file["processed/process/"]:
                mask = data_file["processed/process/mask"][()]
                self.configuration.mask_model.set_dimension(mask.shape)
                self.configuration.mask_model.set_mask(mask)

            if "mask_file" in data_file["processed/process/"]:
                try:
                    self.used_mask = str(data_file["processed/process/mask_file"][()])
                    mask_data = np.array(Image.open(self.used_mask))
                    self.configuration.mask_model.set_dimension(mask_data.shape)
                    self.configuration.mask_model.load_mask(self.used_mask)
                except FileNotFoundError:
                    logger.info(f"Mask file {self.used_mask} is not found")

            if "bkg" in data_file["processed/process/"]:
                self.bkg = data_file["processed/process/bkg"][()]

    def save_proc_data(self, filename):
        """
        Save diffraction patterns to h5 file
        """
        if os.path.dirname(filename) != "":
            os.makedirs(os.path.dirname(filename), exist_ok=True)
        with h5py.File(filename, mode="w") as f:
            f.attrs["default"] = "processed"

            nxentry = f.create_group("processed")
            nxentry.attrs["NX_class"] = "NXentry"
            nxentry.attrs["default"] = "result"

            nxdata = nxentry.create_group("result")
            nxdata.attrs["NX_class"] = "NXdata"
            nxdata.attrs["signal"] = "data"
            nxdata.attrs["axes"] = [".", "binning"]

            nxprocess = nxentry.create_group("process")
            nxprocess.attrs["NX_class"] = "NXprocess"

            if self.used_calibration is not None:
                nxprocess["cal_file"] = str(self.used_calibration)

            if self.used_mask is not None:
                nxprocess["mask_file"] = str(self.used_mask)
                nxprocess["mask_shape"] = self.used_mask_shape

            nxprocess["int_method"] = "csr"
            nxprocess["int_unit"] = "2th_deg"
            nxprocess["num_points"] = self.binning.shape[0]

            if self.bkg is not None:
                nxprocess.create_dataset("bkg", data=self.bkg)

            nxdata.create_dataset("data", data=self.data)
            tth = nxdata.create_dataset("binning", data=self.binning)
            tth.attrs["unit"] = "deg"
            tth.attrs["long_name"] = "two_theta (degrees)"

            nxprocess.create_dataset("pos_map", data=self.pos_map)
            nxprocess.create_dataset("file_map", data=self.file_map)
            nxprocess.create_dataset("files", data=self.files.astype("S"))

    def save_as_csv(self, filename):
        """
        Save diffraction patterns to 3-columns csv file
        """
        if os.path.dirname(filename) != "":
            os.makedirs(os.path.dirname(filename), exist_ok=True)
        x = self.binning.repeat(self.n_img)
        y = (
            np.arange(self.n_img)[None, :]
            .repeat(self.binning.shape[0], axis=0)
            .flatten()
        )
        np.savetxt(
            filename,
            np.array(list(zip(x, y, self.data.T.flatten()))),
            delimiter=",",
            fmt="%f",
        )

    def integrate_raw_data(self, start, stop, step, use_all=False, callback_fn=None):
        """
        Integrate images from given file

        :param start: Start image index from integration
        :param stop: Stop image index from integration
        :param step: Step along images to integrate
        :param use_all: Use all images. If False use only images, that were already integrated.
        :param callback_fn: callback function which is called each iteration with the current image number as parameter,
                            if it returns False the integration will be aborted.
        """
        if self.configuration.use_mask:
            if self.configuration.mask_model.filename != "":
                self.used_mask = self.configuration.mask_model.filename
            mask = self.configuration.mask_model.get_mask()
            self.used_mask_shape = mask.shape

        cal = self.configuration.calibration_model
        unit = self.configuration.integration_unit
        azi_range = self.configuration.oned_azimuth_range

        if cal.can_use_dioptrin_batch(unit, azi_range):
            self._integrate_raw_data_dioptrin_batch(
                start, stop, step, use_all, callback_fn
            )
        else:
            self._integrate_raw_data_pyFAI(start, stop, step, use_all, callback_fn)

    def _integrate_raw_data_pyFAI(
        self, start, stop, step, use_all=False, callback_fn=None
    ):
        intensity_data = []
        binning_data = []
        pos_map = []
        image_counter = 0
        current_file = ""

        # Block img_changed to prevent auto-integration and GUI updates
        self.configuration.img_model.img_changed.blocked = True
        last_callback_time = time.monotonic()
        try:
            for index in range(start, stop, step):
                if use_all:
                    file_index, pos = self.pos_map_all[index]
                else:
                    file_index, pos = self.pos_map[index]
                if file_index != current_file:
                    current_file = file_index
                    self.configuration.calibration_model.img_model.load(
                        self.files[file_index]
                    )
                    self.configuration.mask_model.set_dimension(
                        self.configuration.img_model.img_data.shape
                    )

                self.configuration.img_model.load_series_img(pos + 1)

                binning, intensity = self.configuration.integrate_image_1d(
                    update_pattern_model=False
                )
                image_counter += 1
                pos_map.append((file_index, pos))
                intensity_data.append(intensity)
                binning_data.append(binning)

                now = time.monotonic()
                if callback_fn is not None and now - last_callback_time > 0.1:
                    last_callback_time = now
                    if not callback_fn(image_counter):
                        break
        finally:
            self.configuration.img_model.img_changed.blocked = False

        # deal with different x lengths due to trimmed zeros:
        binning_lengths = [len(b) for b in binning_data]
        binning_max_length_ind = np.argmax(binning_lengths)
        binning_max_length = binning_lengths[binning_max_length_ind]
        binning = binning_data[binning_max_length_ind]

        padded_data = np.zeros((len(intensity_data), binning_max_length))
        for ind, intensity in enumerate(intensity_data):
            padded_data[ind, : len(intensity)] = intensity

        # finish and save everything

        if self.configuration.calibration_model.filename != "":
            self.used_calibration = self.configuration.calibration_model.filename
        self.pos_map = np.array(pos_map)
        self.binning = np.array(binning)
        self.data = padded_data
        self.bkg = None
        self.n_img = self.data.shape[0]

    def _integrate_raw_data_dioptrin_batch(
        self, start, stop, step, use_all=False, callback_fn=None
    ):
        """Load frames through ImgModel, integrate via batch1d_iter with generator."""
        from dioptas.model.util.integration import iter_frames_indexed

        cal = self.configuration.calibration_model
        unit = self.configuration.integration_unit
        num_points = self.configuration.integration_rad_points

        if self.configuration.use_mask:
            mask = self.configuration.mask_model.get_mask()
        elif self.configuration.mask_model.roi is not None:
            mask = self.configuration.mask_model.roi_mask
        else:
            mask = None

        indices = list(range(start, stop, step))
        source = self.pos_map_all if use_all else self.pos_map

        # Configure integrator: load one image to get shape
        first_file_index = source[indices[0]][0]
        self.configuration.img_model.img_changed.blocked = True
        try:
            self.configuration.calibration_model.img_model.load(
                self.files[first_file_index]
            )
            self.configuration.mask_model.set_dimension(
                self.configuration.img_model.img_data.shape
            )
            img_shape = self.configuration.img_model.img_data.shape
            num_points = cal.sync_dioptrin_for_batch(mask, unit, num_points, img_shape)

            # Build pos_map for all indices
            all_pos_map = [(source[i][0], source[i][1]) for i in indices]

            intensity_data = []
            pos_map = []
            binning = None
            aborted = False

            def frame_generator():
                yield from iter_frames_indexed(
                    self.configuration.img_model,
                    self.files,
                    source,
                    indices,
                    mask_model=self.configuration.mask_model,
                    abort_check=lambda: aborted,
                )

            result_iter = cal.dioptrin_batch1d_iter(frame_generator(), num_points)

            last_callback_time = time.monotonic()
            for i, result in enumerate(result_iter):
                if not result.is_ok():
                    raise RuntimeError(
                        f"Dioptrin batch integration failed: {result.error}"
                    )

                x = np.array(result.result.radial)
                y = np.array(result.result.intensity)

                if binning is None:
                    binning = x
                intensity_data.append(y)
                pos_map.append(all_pos_map[i])

                now = time.monotonic()
                if callback_fn is not None and now - last_callback_time > 0.1:
                    last_callback_time = now
                    if not callback_fn(i + 1):
                        aborted = True
                        break
        finally:
            self.configuration.img_model.img_changed.blocked = False

        self._finalize_batch_results(cal, intensity_data, pos_map, binning, unit)

    def set_integration_results(self, results: dict):
        """Apply pre-computed integration results.

        Accepts both pyFAI-style results (``binning_data`` with variable-length
        arrays) and dioptrin-style results (``binning`` with a single array).
        """
        intensity_data = results["intensity_data"]
        pos_map = results["pos_map"]

        if "binning_data" in results:
            # pyFAI path: variable-length arrays, need padding
            binning_data = results["binning_data"]
            binning_lengths = [len(b) for b in binning_data]
            max_len_ind = int(np.argmax(binning_lengths))
            max_len = binning_lengths[max_len_ind]
            binning = binning_data[max_len_ind]

            padded = np.zeros((len(intensity_data), max_len))
            for i, intensity in enumerate(intensity_data):
                padded[i, : len(intensity)] = intensity
            intensity_data = padded
        else:
            # dioptrin path: uniform-length arrays
            binning = results["binning"]
            intensity_data = np.array(intensity_data)

        self.pos_map = np.array(pos_map)
        self.binning = np.array(binning)
        self.data = intensity_data
        self.bkg = None
        self.n_img = self.data.shape[0]

    def _finalize_batch_results(self, cal, intensity_data, pos_map, binning, unit):
        """Store batch integration results."""
        from dioptas.model.util.integration import convert_tth_to_d

        if not intensity_data:
            return

        if unit == "d_A":
            binning = convert_tth_to_d(binning, cal.pattern_geometry.wavelength)

        if self.configuration.calibration_model.filename != "":
            self.used_calibration = self.configuration.calibration_model.filename
        self.pos_map = np.array(pos_map)
        self.binning = np.array(binning)
        self.data = np.array(intensity_data)
        self.bkg = None
        self.n_img = self.data.shape[0]

    def extract_background(self, parameters, callback_fn=None):
        """
        Subtract background calculated with respect of given parameters
        """

        bkg = np.zeros(self.data.shape)
        auto_bkg = SmoothBrucknerBackground(*parameters)
        for i, y in enumerate(self.data):
            if callback_fn is not None:
                if not callback_fn(i):
                    break
            bkg[i] = auto_bkg.extract_background(Pattern(self.binning, y))
        self.bkg = bkg

    def normalize(self, range_ind=(10, 30)):
        if self.data is None:
            return
        average_intensities = np.mean(self.data[:, range_ind[0] : range_ind[1]], axis=1)
        factors = average_intensities[0] / average_intensities
        self.data = (self.data.T * factors).T

    def get_image_info(self, index, use_all=False):
        """
        Get filename and image position in the file

        :param index: Index of image in the batch
        :param use_all: Indexing with respect to all images. If False count only images, that were integrated.
        """
        if use_all:
            if not self.raw_available:
                return None, None
            f_index, pos = self.pos_map_all[index]
        else:
            if self.pos_map is None:
                return "NA", index
            f_index, pos = self.pos_map[index]
        filename = self.files[f_index]
        return filename, pos

    def load_image(self, index, use_all=False):
        """
        Load image in image model

        :param index: Index of image in the batch
        :param use_all: Indexing with respect to all images. If False count only images, that were integrated.
        """
        if not self.raw_available:
            return
        filename, pos = self.get_image_info(index, use_all)
        self.configuration.calibration_model.img_model.load(filename, pos)

    def get_next_folder_filenames(self):
        """
        Loads all files from the next folder with similar file-endings.
        """
        folder_path, _ = os.path.split(self.files[0])
        next_folder_path = iterate_folder(folder_path, 1)
        files = []
        if next_folder_path is not None and os.path.exists(next_folder_path):
            for file in os.listdir(next_folder_path):
                if file.endswith(pathlib.Path(self.files[0]).suffix):
                    files.append(os.path.join(next_folder_path, file))
        files = sorted(files)
        return files[: self.n_img_all]

    def get_previous_folder_filenames(self):
        """
        Loads all files from the previous folder with similar file-endings.
        """
        folder_path, _ = os.path.split(self.files[0])
        previous_folder_path = iterate_folder(folder_path, -1)
        files = []
        if previous_folder_path is not None and os.path.exists(previous_folder_path):
            for file in os.listdir(previous_folder_path):
                if file.endswith(pathlib.Path(self.files[0]).suffix):
                    files.append(os.path.join(previous_folder_path, file))
        files = sorted(files)
        return files[: self.n_img_all]


def iterate_folder(folder_path, step):
    pattern = re.compile(r"\d+")
    match_iterator = pattern.finditer(folder_path)
    new_directory_str = None
    for ind, match in enumerate(list(match_iterator)):
        number_span = match.span()
        left_ind = number_span[0]
        right_ind = number_span[1]
        number = int(folder_path[left_ind:right_ind]) + step
        if number < 0:
            number = 0
        new_directory_str = "{left_str}{number:0{len}}{right_str}".format(
            left_str=folder_path[:left_ind],
            number=number,
            len=right_ind - left_ind,
            right_str=folder_path[right_ind:],
        )
    return new_directory_str



# SPDX-License-Identifier: MIT

from .CalibrationModel import CalibrationModel
from .ImgModel import ImgModel
from .MaskModel import MaskModel
from .PhaseModel import PhaseModel
from .PatternModel import PatternModel
from .OverlayModel import OverlayModel
from .MapModel import MapModel
from .BatchModel import BatchModel



# SPDX-License-Identifier: MIT
import pytest

from ..utility import QtTest
import os
import gc

from mock import MagicMock

from ...controller.integration import PhaseInCakeController
from ...model.DioptasModel import DioptasModel
from ...widgets.integration import IntegrationWidget

unittest_path = os.path.dirname(__file__)
data_path = os.path.join(unittest_path, '../data')
jcpds_path = os.path.join(data_path, 'jcpds')


@pytest.fixture
def phase_in_cake_controller(integration_widget, dioptas_model):
    return PhaseInCakeController(integration_widget, dioptas_model)


@pytest.fixture
def load_phase(phase_in_cake_controller):
    def _load_phase(filename):
        phase_in_cake_controller.model.phase_model.add_jcpds(os.path.join(jcpds_path, filename))

    return _load_phase


@pytest.fixture
def load_phases(load_phase):
    load_phase('ar.jcpds')
    load_phase('ag.jcpds')
    load_phase('au_Anderson.jcpds')
    load_phase('mo.jcpds')
    load_phase('pt.jcpds')
    load_phase('re.jcpds')


def test_loading_a_phase(load_phase, integration_widget):
    assert len(integration_widget.cake_widget.phases) == 0
    load_phase('ar.jcpds')
    assert len(integration_widget.cake_widget.phases) == 1
    load_phase('ar.jcpds')


def test_loading_many_phases(load_phases, integration_widget):
    assert len(integration_widget.cake_widget.phases) == 6


class PhaseInCakeControllerTest(QtTest):
    # SETUP
    #######################
    def setUp(self) -> None:
        self.model = DioptasModel()
        self.model.calibration_model.is_calibrated = True
        self.model.calibration_model.pattern_geometry.wavelength = 0.31E-10
        self.model.calibration_model.integrate_1d = MagicMock(return_value=(self.model.calibration_model.tth,
                                                                            self.model.calibration_model.int))
        self.widget = IntegrationWidget()
        self.controller = PhaseInCakeController(self.widget, self.model)

    def tearDown(self) -> None:
        del self.controller
        del self.widget
        self.model.delete_configurations()
        del self.model
        gc.collect()

    # Utility Functions
    #######################
    def load_phase(self, filename):
        self.model.phase_model.add_jcpds(os.path.join(jcpds_path, filename))

    def load_phases(self):
        self.load_phase('ar.jcpds')
        self.load_phase('ag.jcpds')
        self.load_phase('au_Anderson.jcpds')
        self.load_phase('mo.jcpds')
        self.load_phase('pt.jcpds')
        self.load_phase('re.jcpds')

    # Tests
    #######################

    def test_remove_phase(self):
        self.load_phases()
        self.assertEqual(len(self.widget.cake_widget.phases), 6)
        self.model.phase_model.del_phase(3)
        self.assertEqual(len(self.widget.cake_widget.phases), 5)

    def test_changing_pressure(self):
        self.load_phase('ar.jcpds')
        first_line_position = self.widget.cake_widget.phases[0].line_items[0].getPos()
        self.model.phase_model.set_pressure(0, 4)
        self.assertNotEqual(first_line_position,
                            self.widget.cake_widget.phases[0].line_items[0].getPos())

    def test_changing_temperature_and_pressure(self):
        self.load_phase('pt.jcpds')
        self.model.phase_model.set_pressure(0, 100)
        first_line_position = self.widget.cake_widget.phases[0].line_items[0].getPos()
        self.model.phase_model.set_temperature(0, 3000)
        self.assertNotEqual(first_line_position,
                            self.widget.cake_widget.phases[0].line_items[0].getPos())

    def test_changing_color(self):
        self.load_phase('pt.jcpds')
        green_value = self.widget.cake_widget.phases[0].line_items[0].pen.color().green()
        self.model.phase_model.set_color(0, (230, 22, 0))
        self.assertNotEqual(green_value,
                            self.widget.cake_widget.phases[0].line_items[0].pen.color().green())

    def test_reflection_added(self):
        self.load_phase("pt.jcpds")
        num_line_items = len(self.widget.cake_widget.phases[0].line_items)

        self.model.phase_model.add_reflection(0)
        self.assertEqual(len(self.widget.cake_widget.phases[0].line_items),
                         num_line_items + 1)

    def test_delete_reflection(self):
        self.load_phase("pt.jcpds")
        num_line_items = len(self.widget.cake_widget.phases[0].line_items)
        self.model.phase_model.delete_reflection(0, 0)
        self.assertEqual(len(self.widget.cake_widget.phases[0].line_items),
                         num_line_items - 1)

    def test_reload_phase(self):
        self.load_phase('au_Anderson.jcpds')
        self.model.phase_model.reload(0)

    def test_phases_are_not_shown_in_img_mode(self):
        self.load_phase('au_Anderson.jcpds')
        cake_phase = self.widget.cake_widget.phases[0]
        self.assertFalse(cake_phase.visible)



# SPDX-License-Identifier: MIT

import os

import pytest

from ...model.PhaseModel import PhaseModel

unittest_path = os.path.dirname(__file__)
data_path = os.path.join(unittest_path, '../data')
jcpds_path = os.path.join(data_path, 'jcpds')


@pytest.fixture
def phase_model():
    return PhaseModel()


def load_phase(phase_model, filename):
    phase_model.add_jcpds(os.path.join(jcpds_path, filename))


def test_same_conditions_set_pressure(phase_model):
    load_phase(phase_model, 'ar.jcpds')
    load_phase(phase_model, 'pt.jcpds')

    assert phase_model.phases[0].params['pressure'] == 0
    assert phase_model.phases[1].params['pressure'] == 0
    assert phase_model.same_conditions

    phase_model.set_pressure(0, 10)

    assert phase_model.phases[0].params['pressure'] == 10
    assert phase_model.phases[1].params['pressure'] == 10

    phase_model.same_conditions = False
    phase_model.set_pressure(1, 5)
    assert phase_model.phases[0].params['pressure'] == 10
    assert phase_model.phases[1].params['pressure'] == 5


def test_same_conditions_set_temperature(phase_model: PhaseModel):
    load_phase(phase_model, 'pt.jcpds')
    load_phase(phase_model, 'pt.jcpds')

    assert phase_model.phases[0].params['temperature'] == 298
    assert phase_model.phases[1].params['temperature'] == 298
    assert phase_model.same_conditions

    phase_model.set_temperature(1, 2000)

    assert phase_model.phases[0].params['temperature'] == 2000
    assert phase_model.phases[1].params['temperature'] == 2000

    phase_model.same_conditions = False
    phase_model.set_temperature(1, 1500)
    assert phase_model.phases[0].params['temperature'] == 2000
    assert phase_model.phases[1].params['temperature'] == 1500


def test_set_temperature_with_no_thermal_expansion(phase_model):
    load_phase(phase_model, 'ar.jcpds')

    assert phase_model.phases[0].params['temperature'] == 298
    assert not phase_model.phases[0].has_thermal_expansion()

    phase_model.set_temperature(0, 2000)

    # since there is no thermal expansion defined the temperature should stay at ambient
    assert phase_model.phases[0].params['temperature'] == 298


def test_reload_phase(phase_model: PhaseModel):
    load_phase(phase_model, 'ar.jcpds')
    num_refl = len(phase_model.reflections[0])
    phase_model.delete_reflection(0, 0)
    phase_model.delete_reflection(0, 0)
    phase_model.set_pressure(0, 5)
    old_a0 = phase_model.phases[0].params['a0']
    phase_model.set_param(0, 'a0', 5)

    phase_model.reload(0)

    assert len(phase_model.reflections[0]) == num_refl
    assert phase_model.phases[0].params['a0'] == old_a0
    assert phase_model.phases[0].params['pressure'] == 5



# SPDX-License-Identifier: MIT

import os.path
from unittest.mock import MagicMock

import fabio
import numpy as np
import pytest
from PIL import Image
from qtpy import QtWidgets

from ..utility import click_button, unittest_data_path
from ...controller.MainController import MainController
from ...controller.MaskController import MaskController


def load_image_and_mask(
    main_controller: MainController,
    img_filename: str,
    mask_filename: str,
    dialog_filter: str,
):
    """Sequence: Load image and mask from file and save the mask to file

    dialog_filter is the format option selected in the mask controller load/save dialogs.
    """
    # Load image
    click_button(main_controller.widget.calibration_mode_btn)
    QtWidgets.QFileDialog.getOpenFileName = MagicMock(return_value=img_filename)
    click_button(main_controller.widget.calibration_widget.load_img_btn)

    # Create mask file
    ref_mask = np.zeros(main_controller.model.mask_model.mask_dimension, dtype=np.int8)
    ref_mask[:100, :200] = 1
    if mask_filename.endswith(".npy"):
        np.save(mask_filename, ref_mask)
    elif mask_filename.endswith(".edf"):
        fabio.edfimage.EdfImage(ref_mask).write(mask_filename)
    else:
        Image.fromarray(ref_mask).save(mask_filename, "tiff")

    # Load mask
    click_button(main_controller.widget.mask_mode_btn)
    QtWidgets.QFileDialog.getOpenFileName = MagicMock(
        return_value=(mask_filename, dialog_filter)
    )
    click_button(main_controller.mask_controller.widget.load_mask_btn)

    current_mask = main_controller.model.mask_model.get_mask()
    if dialog_filter.startswith(MaskController.FLIPUD_MASK_FILTER_PREFIX):
        assert np.array_equal(np.flipud(ref_mask), current_mask)
    else:
        assert np.array_equal(ref_mask, current_mask)

    # Save mask
    QtWidgets.QFileDialog.getSaveFileName = MagicMock(
        return_value=(mask_filename, dialog_filter)
    )
    click_button(main_controller.mask_controller.widget.save_mask_btn)
    assert os.path.isfile(mask_filename)

    if mask_filename.endswith(".npy"):
        saved_mask = np.load(mask_filename)
    elif mask_filename.endswith(".edf"):
        saved_mask = fabio.open(mask_filename).data
    else:
        saved_mask = np.array(Image.open(mask_filename))
    assert np.array_equal(ref_mask, saved_mask)


@pytest.mark.parametrize(
    "img_filename",
    [
        "lambda/testasapo1_1009_00002_m1_part00000.nxs",
        "spe/CeO2_PI_CCD_Mo.SPE",
        "image_001.tif",
        "karabo_epix.h5",
    ],
)
def test_load_save_mask_as_tiff(main_controller, tmp_path, img_filename):
    """Test *.mask mask load/save"""
    load_image_and_mask(
        main_controller,
        os.path.join(unittest_data_path, img_filename),
        str(tmp_path / "ref_mask.mask"),
        dialog_filter=MaskController.DEFAULT_MASK_FILTER,
    )


@pytest.mark.parametrize(
    "img_filename",
    [
        "lambda/testasapo1_1009_00002_m1_part00000.nxs",
        "spe/CeO2_PI_CCD_Mo.SPE",
        "image_001.tif",
        "karabo_epix.h5",
    ],
)
def test_load_save_flipped_mask_as_npy(main_controller, tmp_path, img_filename):
    """Test load/save flipped mask as npy"""
    load_image_and_mask(
        main_controller,
        os.path.join(unittest_data_path, img_filename),
        str(tmp_path / "ref_mask.npy"),
        dialog_filter=f"{MaskController.FLIPUD_MASK_FILTER_PREFIX} (*.npy)",
    )


@pytest.mark.parametrize(
    "img_filename",
    [
        "lambda/testasapo1_1009_00002_m1_part00000.nxs",
        "spe/CeO2_PI_CCD_Mo.SPE",
        "image_001.tif",
        "karabo_epix.h5",
    ],
)
def test_load_save_flipped_mask_as_edf(main_controller, tmp_path, img_filename):
    """Test load/save flipped mask as edf"""
    load_image_and_mask(
        main_controller,
        os.path.join(unittest_data_path, img_filename),
        str(tmp_path / "ref_mask.edf"),
        dialog_filter=f"{MaskController.FLIPUD_MASK_FILTER_PREFIX} (*.edf)",
    )



# SPDX-License-Identifier: MIT
from __future__ import absolute_import

import sys
import os
import time

try:
    from cStringIO import StringIO
except ImportError:
    from io import StringIO
import traceback
from qtpy import QtWidgets


def excepthook(exc_type, exc_value, traceback_obj):
    """
    Global function to catch unhandled exceptions. This function will result in an error dialog which displays the
    error information.

    :param exc_type: exception type
    :param exc_value: exception value
    :param traceback_obj: traceback object
    :return:
    """

    separator = '-' * 80
    tb_info_file = StringIO()
    traceback.print_tb(traceback_obj, None, tb_info_file)
    traceback.print_exception(exc_type, exc_value, traceback_obj)
    tb_info_file.seek(0)
    tb_info = tb_info_file.read()
    errmsg = '%s: \n%s' % (str(exc_type), str(exc_value))
    sections = [separator, errmsg, separator, tb_info]
    msg = '\n'.join(sections)

    print(msg)


def main():
    sys.excepthook = excepthook


# SPDX-License-Identifier: MIT

from ..utility import QtTest
import os
import gc

from qtpy import QtWidgets
from mock import MagicMock, patch

from ..utility import enter_value_into_text_field, click_button

from ...controller.integration import OptionsController
from ...model.DioptasModel import DioptasModel
from ...widgets.integration import IntegrationWidget

unittest_path = os.path.dirname(__file__)
data_path = os.path.join(unittest_path, '../data')


class OptionsControllerTest(QtTest):
    def setUp(self):
        # Mocking the function which will block the unittest for some reason...
        # Use patch to ensure proper cleanup
        self.processEvents_patcher = patch.object(QtWidgets.QApplication, 'processEvents', MagicMock())
        self.processEvents_patcher.start()

        self.widget = IntegrationWidget()
        self.options_widget = self.widget.integration_control_widget.integration_options_widget
        self.model = DioptasModel()

        self.options_controller = OptionsController(self.widget, self.model)

    def tearDown(self):
        # Stop the patcher to restore original behavior
        self.processEvents_patcher.stop()

        # Clean up widgets
        self.widget.close()
        self.widget.deleteLater()
        del self.options_controller
        del self.widget
        del self.model
        gc.collect()

    def test_change_azimuth_bins(self):
        enter_value_into_text_field(self.options_widget.cake_azimuth_points_sb.lineEdit(), 100)
        self.assertEqual(self.model.current_configuration.cake_azimuth_points, 100)

    def test_change_azimuth_range(self):
        click_button(self.options_widget.cake_full_toggle_btn)
        enter_value_into_text_field(self.options_widget.cake_azimuth_min_txt, -100)
        self.assertEqual(self.model.current_configuration.cake_azimuth_range[0], -100)

        enter_value_into_text_field(self.options_widget.cake_azimuth_max_txt, 200)
        self.assertEqual(self.model.current_configuration.cake_azimuth_range[1], 200)






# SPDX-License-Identifier: MIT

from ..utility import QtTest, click_button, click_checkbox
import pytest
import os
import gc
from mock import MagicMock

import numpy as np
from qtpy import QtWidgets, QtCore, QtGui
from qtpy.QtTest import QTest

from ...controller.integration import OverlayController
from ...model.DioptasModel import DioptasModel
from ...model.OverlayModel import OverlayModel
from ...widgets.integration import IntegrationWidget

unittest_path = os.path.dirname(__file__)
data_path = os.path.join(unittest_path, "../data")


@pytest.fixture
def overlay_controller(
    integration_widget: IntegrationWidget, dioptas_model: DioptasModel
) -> OverlayController:
    return OverlayController(integration_widget, dioptas_model)


def load_overlays(overlay_model: OverlayModel):
    for i in range(6):
        overlay_model.add_overlay(np.arange(10), np.arange(10), "Overlay {}".format(i))


def test_add_overlays(
    overlay_controller: OverlayController,
    integration_widget: IntegrationWidget,
    dioptas_model: DioptasModel,
):
    load_overlays(dioptas_model.overlay_model)

    overlay_tw = integration_widget.overlay_widget.overlay_tw

    assert overlay_tw.rowCount() == 6
    assert len(dioptas_model.overlay_model.overlays) == 6
    assert overlay_tw.currentRow() == 5


def test_delete_overlays(
    overlay_controller: OverlayController,
    integration_widget: IntegrationWidget,
    dioptas_model: DioptasModel,
):
    load_overlays(dioptas_model.overlay_model)

    overlay_tw = integration_widget.overlay_widget.overlay_tw
    overlay_widget = integration_widget.overlay_widget

    overlay_controller.delete_btn_click_callback()

    assert overlay_tw.rowCount() == 5
    assert len(dioptas_model.overlay_model.overlays) == 5
    assert overlay_tw.currentRow() == 4

    overlay_widget.select_overlay(1)
    overlay_controller.delete_btn_click_callback()

    assert overlay_tw.rowCount() == 4
    assert len(dioptas_model.overlay_model.overlays) == 4
    assert overlay_tw.currentRow() == 1

    tw_filenames = [overlay_tw.item(i, 2).text() for i in range(4)]
    assert tw_filenames[1] == "Overlay 2"

    for _ in range(4):
        click_button(overlay_widget.delete_btn)

    assert overlay_tw.rowCount() == 0
    assert len(dioptas_model.overlay_model.overlays) == 0
    assert overlay_tw.currentRow() == -1

    click_button(overlay_widget.delete_btn)
    assert overlay_tw.rowCount() == 0
    assert len(dioptas_model.overlay_model.overlays) == 0
    assert overlay_tw.currentRow() == -1


def test_clear_overlays(
    overlay_controller: OverlayController,
    integration_widget: IntegrationWidget,
    dioptas_model: DioptasModel,
):
    load_overlays(dioptas_model.overlay_model)

    overlay_tw = integration_widget.overlay_widget.overlay_tw
    overlay_widget = integration_widget.overlay_widget

    click_button(overlay_widget.clear_btn)

    assert overlay_tw.rowCount() == 0
    assert len(dioptas_model.overlay_model.overlays) == 0
    assert overlay_tw.currentRow() == -1


def test_change_scaling_in_view(
    overlay_controller: OverlayController,
    integration_widget: IntegrationWidget,
    dioptas_model: DioptasModel,
):
    load_overlays(dioptas_model.overlay_model)
    overlay_widget = integration_widget.overlay_widget
    for ind in [0, 3, 4]:
        overlay_widget.scale_sbs[ind].setValue(2.0)
        assert dioptas_model.overlay_model.get_overlay_scaling(ind) == 2


def test_change_offset_in_view(
    overlay_controller: OverlayController,
    integration_widget: IntegrationWidget,
    dioptas_model: DioptasModel,
):
    load_overlays(dioptas_model.overlay_model)
    overlay_widget = integration_widget.overlay_widget
    for ind in [0, 3, 4]:
        overlay_widget.offset_sbs[ind].setValue(100)
        assert dioptas_model.overlay_model.get_overlay_offset(ind) == 100


def test_scaling_auto_step_change(
    overlay_controller: OverlayController,
    integration_widget: IntegrationWidget,
    dioptas_model: DioptasModel,
):
    load_overlays(dioptas_model.overlay_model)
    overlay_widget = integration_widget.overlay_widget

    overlay_widget.scale_step_msb.setValue(0.5)
    overlay_widget.scale_step_msb.stepUp()

    new_scale_step = overlay_widget.scale_step_msb.value()
    assert new_scale_step == pytest.approx(1.0)

    overlay_widget.scale_step_msb.stepDown()
    overlay_widget.scale_step_msb.stepDown()
    new_scale_step = overlay_widget.scale_step_msb.value()
    assert new_scale_step == pytest.approx(0.2)


def test_scalestep_spinbox_changes_scale_spinboxes(
    overlay_controller: OverlayController,
    integration_widget: IntegrationWidget,
    dioptas_model: DioptasModel,
):
    overlay_widget = integration_widget.overlay_widget
    load_overlays(dioptas_model.overlay_model)
    for ind in range(6):
        assert integration_widget.overlay_widget.scale_sbs[ind].singleStep() == 0.01

    new_step = 5
    overlay_widget.scale_step_msb.setValue(new_step)
    for ind in range(6):
        assert integration_widget.overlay_widget.scale_sbs[ind].singleStep() == new_step


def test_offset_spinbox_changes_offset_spinboxes(
    overlay_controller: OverlayController,
    integration_widget: IntegrationWidget,
    dioptas_model: DioptasModel,
):
    overlay_widget = integration_widget.overlay_widget
    load_overlays(dioptas_model.overlay_model)
    for ind in range(6):
        assert overlay_widget.offset_sbs[ind].singleStep() == 100

    new_step = 5
    overlay_widget.offset_step_msb.setValue(new_step)
    for ind in range(6):
        assert overlay_widget.offset_sbs[ind].singleStep() == new_step


def test_offset_auto_step_change(
    overlay_controller: OverlayController,
    integration_widget: IntegrationWidget,
    dioptas_model: DioptasModel,
):
    overlay_widget = integration_widget.overlay_widget
    load_overlays(dioptas_model.overlay_model)
    overlay_widget.offset_step_msb.setValue(10.0)
    overlay_widget.offset_step_msb.stepUp()

    new_offset_step = overlay_widget.offset_step_msb.value()
    assert new_offset_step == pytest.approx(20.0)

    overlay_widget.offset_step_msb.stepDown()
    overlay_widget.offset_step_msb.stepDown()
    new_offset_step = overlay_widget.offset_step_msb.value()
    assert new_offset_step == pytest.approx(5.0)


def test_setting_overlay_as_bkg(
    overlay_controller: OverlayController,
    integration_widget: IntegrationWidget,
    dioptas_model: DioptasModel,
):
    overlay_widget = integration_widget.overlay_widget
    load_overlays(dioptas_model.overlay_model)
    dioptas_model.pattern_model.set_pattern(np.arange(10), np.arange(10))
    overlay_widget.select_overlay(0)

    click_button(integration_widget.overlay_set_as_bkg_btn)

    assert integration_widget.overlay_set_as_bkg_btn.isChecked()
    assert np.array_equal(
        dioptas_model.pattern_model.background_pattern,
        dioptas_model.overlay_model.overlays[0],
    )
    _, y = dioptas_model.pattern.data
    assert np.sum(y) == 0


def test_setting_overlay_as_bkg_and_changing_scale(
    overlay_controller: OverlayController,
    integration_widget: IntegrationWidget,
    dioptas_model: DioptasModel,
):
    overlay_widget = integration_widget.overlay_widget
    load_overlays(dioptas_model.overlay_model)
    dioptas_model.pattern_model.set_pattern(np.arange(10), np.arange(10))
    overlay_widget.select_overlay(0)

    click_button(integration_widget.overlay_set_as_bkg_btn)

    overlay_widget.scale_sbs[0].setValue(2)

    _, y = dioptas_model.pattern.data
    assert np.array_equal(y, -np.arange(10))


def test_setting_overlay_as_bkg_and_changing_offset(
    overlay_controller: OverlayController,
    integration_widget: IntegrationWidget,
    dioptas_model: DioptasModel,
):
    overlay_widget = integration_widget.overlay_widget
    load_overlays(dioptas_model.overlay_model)

    dioptas_model.pattern_model.set_pattern(np.arange(10), np.arange(10))
    overlay_widget.select_overlay(0)

    click_button(integration_widget.overlay_set_as_bkg_btn)

    overlay_widget.offset_sbs[0].setValue(100)
    _, y = dioptas_model.pattern.data
    assert np.sum(y) == -100 * y.size


def test_setting_overlay_as_bkg_and_then_change_to_new_overlay_as_bkg(
    overlay_controller: OverlayController,
    integration_widget: IntegrationWidget,
    dioptas_model: DioptasModel,
):
    overlay_widget = integration_widget.overlay_widget
    load_overlays(dioptas_model.overlay_model)
    dioptas_model.pattern_model.set_pattern(np.arange(10), np.arange(10))
    overlay_widget.select_overlay(0)

    click_button(integration_widget.overlay_set_as_bkg_btn)

    assert integration_widget.overlay_set_as_bkg_btn.isChecked()
    assert (
        dioptas_model.pattern.background_pattern is
        dioptas_model.overlay_model.overlays[0]
    )
    _, y = dioptas_model.pattern.data
    assert np.sum(y) == 0

    overlay_widget.select_overlay(1)
    assert not integration_widget.overlay_set_as_bkg_btn.isChecked()
    overlay_widget.scale_sbs[1].setValue(2)
    click_button(integration_widget.overlay_set_as_bkg_btn)

    assert (
        dioptas_model.pattern.background_pattern is
        dioptas_model.overlay_model.overlays[1]
    )
    _, y = dioptas_model.pattern.data
    assert np.array_equal(y, -np.arange(10))


def test_setting_pattern_as_bkg(
    overlay_controller: OverlayController,
    integration_widget: IntegrationWidget,
    dioptas_model: DioptasModel,
):
    dioptas_model.pattern_model.set_pattern(np.arange(10), np.arange(10))
    click_button(integration_widget.qa_set_as_background_btn)

    assert integration_widget.overlay_set_as_bkg_btn.isChecked()
    assert len(dioptas_model.overlay_model.overlays) == 1

    _, y = dioptas_model.pattern.data
    assert np.sum(y) == 0


def test_having_overlay_as_bkg_and_deleting_it(
    overlay_controller: OverlayController,
    integration_widget: IntegrationWidget,
    dioptas_model: DioptasModel,
):
    dioptas_model.pattern_model.set_pattern(np.arange(10), np.arange(10))
    click_button(integration_widget.qa_set_as_background_btn)

    assert dioptas_model.pattern_model.background_pattern is not None
    assert len(dioptas_model.overlay_model.overlays) == 1

    click_button(integration_widget.overlay_del_btn)

    assert not integration_widget.overlay_set_as_bkg_btn.isChecked()
    assert integration_widget.overlay_tw.rowCount() == 0
    assert len(dioptas_model.overlay_model.overlays) == 0

    _, y = dioptas_model.pattern.data
    assert np.sum(y) != 0


def test_overlay_waterfall(
    overlay_controller: OverlayController,
    integration_widget: IntegrationWidget,
    dioptas_model: DioptasModel,
):
    overlay_widget = integration_widget.overlay_widget
    load_overlays(dioptas_model.overlay_model)
    overlay_widget.waterfall_separation_msb.setValue(10)
    click_button(overlay_widget.waterfall_btn)

    assert dioptas_model.overlay_model.overlays[5].offset == -10
    assert dioptas_model.overlay_model.overlays[4].offset == -20

    click_button(integration_widget.reset_waterfall_btn)

    assert dioptas_model.overlay_model.overlays[5].offset == 0
    assert dioptas_model.overlay_model.overlays[5].offset == 0


def test_move_single_overlay_one_step_up(
    overlay_controller: OverlayController,
    integration_widget: IntegrationWidget,
    dioptas_model: DioptasModel,
):
    overlay_model = dioptas_model.overlay_model
    load_overlays(dioptas_model.overlay_model)
    integration_widget.overlay_tw.selectRow(3)

    assert integration_widget.overlay_tw.item(3, 2).text() == "Overlay 3"
    assert overlay_model.overlays[2].name == "Overlay 2"

    new_name = "special"
    overlay_model.set_overlay_name(3, new_name)
    assert overlay_model.overlays[3].name == new_name
    assert integration_widget.overlay_tw.item(3, 2).text() == new_name

    click_button(integration_widget.overlay_move_up_btn)

    assert dioptas_model.overlay_model.overlays[2].name == new_name
    assert integration_widget.overlay_tw.item(2, 2).text() == new_name
    assert dioptas_model.overlay_model.overlays[3].name == "Overlay 2"
    assert integration_widget.overlay_tw.item(3, 2).text() == "Overlay 2"

    # selected row should also move
    assert integration_widget.overlay_tw.currentRow() == 2


def test_move_single_overlay_one_step_down(
    overlay_controller: OverlayController,
    integration_widget: IntegrationWidget,
    dioptas_model: DioptasModel,
):
    overlay_model = dioptas_model.overlay_model
    load_overlays(dioptas_model.overlay_model)
    integration_widget.overlay_tw.selectRow(3)

    assert integration_widget.overlay_tw.item(3, 2).text() == "Overlay 3"
    assert overlay_model.overlays[2].name == "Overlay 2"

    new_name = "special"
    overlay_model.set_overlay_name(3, new_name)
    assert overlay_model.overlays[3].name == new_name
    assert integration_widget.overlay_tw.item(3, 2).text() == new_name

    click_button(integration_widget.overlay_move_down_btn)

    assert dioptas_model.overlay_model.overlays[4].name == new_name
    assert integration_widget.overlay_tw.item(4, 2).text() == new_name
    assert dioptas_model.overlay_model.overlays[3].name == "Overlay 4"
    assert integration_widget.overlay_tw.item(3, 2).text() == "Overlay 4"

    # selected row should also move
    assert integration_widget.overlay_tw.currentRow() == 4


def test_bulk_change_visibility_of_overlays(
    overlay_controller: OverlayController,
    integration_widget: IntegrationWidget,
    dioptas_model: DioptasModel,
):
    overlay_widget = integration_widget.overlay_widget
    load_overlays(dioptas_model.overlay_model)
    for cb in overlay_widget.show_cbs:
        assert cb.isChecked()

    overlay_controller.overlay_tw_header_section_clicked(0)
    for cb in overlay_widget.show_cbs:
        assert not cb.isChecked()

    for overlay in dioptas_model.overlay_model.overlays:
        assert not overlay.visible

    click_checkbox(overlay_widget.show_cbs[1])
    overlay_controller.overlay_tw_header_section_clicked(0)
    for cb in overlay_widget.show_cbs:
        assert not cb.isChecked()

    overlay_controller.overlay_tw_header_section_clicked(0)
    for cb in overlay_widget.show_cbs:
        assert cb.isChecked()

    for overlay in dioptas_model.overlay_model.overlays:
        assert overlay.visible


def test_change_overlay_color(
    overlay_controller: OverlayController,
    integration_widget: IntegrationWidget,
    dioptas_model: DioptasModel,
):
    overlay_widget = integration_widget.overlay_widget
    load_overlays(dioptas_model.overlay_model)
    color1 = overlay_widget.color_btns[1].palette().color(QtGui.QPalette.Button)
    QtWidgets.QColorDialog.getColor = MagicMock(return_value=color1)

    click_button(overlay_widget.color_btns[3])

    assert dioptas_model.overlay_model.overlays[3].color == color1.name()
    assert overlay_widget.color_btns[3].palette().color(QtGui.QPalette.Button) == color1



# SPDX-License-Identifier: MIT

#x - horizontal motor
#y - vertical motor
#z - focus

epics_config = {
    'sample_position_x': '13IDD:m81',
    'sample_position_y': '13IDD:m83',
    'sample_position_z': '13IDD:m82',
    'sample_position_omega': '13IDD:m96'}


# SPDX-License-Identifier: MIT

import os

import pytest

from ...model.util.HelperModule import FileNameIterator

unittest_path = os.path.dirname(__file__)
data_path = os.path.join(unittest_path, '../data')


@pytest.fixture
def filename_iterator():
    return FileNameIterator()


def test_get_next_filename_with_existent_file(filename_iterator):
    filename = 'image_001.tif'
    filename_iterator.update_filename(os.path.join(data_path, filename))
    new_filename = os.path.basename(filename_iterator.get_next_filename())
    assert new_filename == 'image_002.tif'


def test_get_next_filename_with_non_existent_file(filename_iterator):
    filename = 'image_002.tif'
    filename_iterator.update_filename(os.path.join(data_path, filename))
    assert filename_iterator.get_next_filename() is None


def test_get_next_filename_with_larger_Step(filename_iterator):
    filename = 'image_000.tif'
    filename_iterator.update_filename(os.path.join(data_path, filename))
    new_filename = os.path.basename(filename_iterator.get_next_filename(step=2))
    assert new_filename == 'image_002.tif'


def test_get_next_filename_with_two_numbers_in_name(filename_iterator, tmp_path):
    filename1 = 'image_001w_001.tif'
    filename2 = 'image_002w_001.tif'
    file_path1 = os.path.join(tmp_path, filename1)
    file_path2 = os.path.join(tmp_path, filename2)

    open(file_path1, "w")
    open(file_path2, "w")

    filename_iterator.update_filename(file_path1)
    new_filename = filename_iterator.get_next_filename()
    assert os.path.abspath(new_filename) == os.path.abspath(file_path2)

    os.remove(file_path1)
    os.remove(file_path2)

    filename1 = 'image_003w_001.tif'
    filename2 = 'image_003w_002.tif'
    filename3 = 'image_004w_001.tif'

    file_path1 = os.path.join(tmp_path, filename1)
    file_path2 = os.path.join(tmp_path, filename2)
    file_path3 = os.path.join(tmp_path, filename3)

    open(file_path1, "w")
    open(file_path2, "w")
    open(file_path3, "w")

    filename_iterator.update_filename(file_path1)
    new_filename = filename_iterator.get_next_filename()
    assert os.path.abspath(new_filename) == os.path.abspath(file_path2)


def test_get_previous_filename_with_existent_file(filename_iterator):
    filename = 'image_002.tif'
    filename_iterator.update_filename(os.path.join(data_path, filename))
    new_filename = os.path.basename(filename_iterator.get_previous_filename())
    assert new_filename == 'image_001.tif'


def test_get_previous_filename_with_non_existent_file(filename_iterator):
    filename = 'image_001.tif'
    filename_iterator.update_filename(os.path.join(data_path, filename))
    assert filename_iterator.get_previous_filename() is None


def test_get_previous_filename_with_larger_Step(filename_iterator):
    filename = 'image_003.tif'
    filename_iterator.update_filename(os.path.join(data_path, filename))
    new_filename = os.path.basename(filename_iterator.get_previous_filename(step=2))
    assert new_filename == 'image_001.tif'



# SPDX-License-Identifier: MIT

import os
import numpy as np
from math import cos, sin

import pytest
from qtpy import QtCore

from ...model.MaskModel import MaskModel

unittest_path = os.path.dirname(__file__)
data_path = os.path.join(unittest_path, '../data')


@pytest.fixture
def mask_model():
    mask_model = MaskModel()
    mask_model.set_dimension((10, 10))
    return mask_model


def test_growing_masks(mask_model):
    mask_model._mask_data[4, 4] = 1
    mask_model._mask_data[0, 0] = 1
    mask_model._mask_data[0, 9] = 1
    mask_model._mask_data[9, 9] = 1
    mask_model._mask_data[9, 0] = 1

    mask_model.grow()

    # tests corners
    assert mask_model._mask_data[0, 1] == 1
    assert mask_model._mask_data[1, 1] == 1
    assert mask_model._mask_data[1, 0] == 1

    assert mask_model._mask_data[0, 8] == 1
    assert mask_model._mask_data[1, 8] == 1
    assert mask_model._mask_data[1, 9] == 1

    assert mask_model._mask_data[8, 0] == 1
    assert mask_model._mask_data[8, 1] == 1
    assert mask_model._mask_data[9, 1] == 1

    assert mask_model._mask_data[8, 8] == 1
    assert mask_model._mask_data[8, 9] == 1
    assert mask_model._mask_data[9, 8] == 1

    # tests center
    assert mask_model._mask_data[3, 3] == 1
    assert mask_model._mask_data[4, 3] == 1
    assert mask_model._mask_data[5, 3] == 1

    assert mask_model._mask_data[3, 5] == 1
    assert mask_model._mask_data[4, 5] == 1
    assert mask_model._mask_data[5, 5] == 1

    assert mask_model._mask_data[3, 4] == 1
    assert mask_model._mask_data[5, 4] == 1


def test_shrink_mask(mask_model):
    mask_model._mask_data[4, 4] = 1
    mask_model._mask_data[0, 0] = 1
    mask_model._mask_data[0, 9] = 1
    mask_model._mask_data[9, 9] = 1
    mask_model._mask_data[9, 0] = 1

    before_mask = np.copy(mask_model._mask_data)
    mask_model.grow()
    mask_model.shrink()

    assert np.array_equal(before_mask, mask_model._mask_data)

    mask_model.clear_mask()

    mask_model._mask_data[4, 4] = 1
    mask_model._mask_data[5, 4] = 1
    mask_model._mask_data[5, 5] = 1
    mask_model._mask_data[4, 5] = 1
    mask_model.shrink()

    assert np.sum(mask_model._mask_data) == 0


@pytest.mark.parametrize("flipud", [False, True])
@pytest.mark.parametrize("extension", [".mask", ".npy", ".edf"])
def test_saving_and_loading(mask_model, tmp_path, extension, flipud):
    mask_model.mask_ellipse(1024, 1024, 100, 100)
    mask_model.set_dimension((2048, 2048))

    mask_array = np.copy(mask_model.get_img())

    filename = os.path.join(tmp_path, f"dummy{extension}")

    mask_model.save_mask(filename, flipud)
    mask_model.load_mask(filename, flipud)

    assert np.array_equal(mask_array, mask_model.get_img())

    mask_model.load_mask(filename, not flipud)
    assert np.array_equal(mask_array, np.flipud(mask_model.get_img()))


def test_use_roi(mask_model):
    mask_model.roi = [0, 2, 0, 2]

    assert np.array_equal(mask_model.get_mask()[0:3, 0:3],
                          np.array([[0, 0, 1],
                                    [0, 0, 1],
                                    [1, 1, 1]]))


@pytest.mark.parametrize("extension", [".mask", ".npy", ".edf"])
def test_save_mask(mask_model, tmp_path, extension):
    mask_model.mask_below_threshold(np.zeros(shape=(10, 10)), 1)
    filename = os.path.join(tmp_path, f"test_save{extension}")
    mask_model.save_mask(filename)

    assert os.path.exists(filename)


def test_find_center_of_circle_from_three_points(mask_model):
    x0 = 2.0
    y0 = 3.5
    r = 1.2
    phi1 = 0.1
    phi2 = 1.3
    phi3 = 6.0
    p1 = QtCore.QPointF(x0 + r * cos(phi1), y0 + r * sin(phi1))
    p2 = QtCore.QPointF(x0 + r * cos(phi2), y0 + r * sin(phi2))
    p3 = QtCore.QPointF(x0 + r * cos(phi3), y0 + r * sin(phi3))
    mask_model.find_center_of_circle_from_three_points(p1, p2, p3)
    assert pytest.approx(x0) == mask_model.center_for_arc.x()
    assert pytest.approx(y0) == mask_model.center_for_arc.y()


def test_find_center_of_circle_from_three_points_collinear(mask_model):
    p1 = QtCore.QPointF(0.0, 0.0)
    p2 = QtCore.QPointF(1.0, 1.0)
    p3 = QtCore.QPointF(2.0, 2.0)
    assert mask_model.find_center_of_circle_from_three_points(p1, p2, p3) is None


def test_find_radius_of_circle_from_center_and_point(mask_model):
    x0 = 2.0
    y0 = 3.5
    p0 = QtCore.QPointF(x0, y0)
    r = 1.2
    phi1 = 0.1
    p1 = QtCore.QPointF(x0 + r * cos(phi1), y0 + r * sin(phi1))
    rcalc = mask_model.find_radius_of_circle_from_center_and_point(p0, p1)
    assert r == rcalc


def test_find_n_points_on_arc_from_three_points(mask_model):
    n = 50
    x0 = 2.0
    y0 = 3.5
    p0 = QtCore.QPointF(x0, y0)
    r = 1.2
    width = 0

    phi1 = 0.1
    phi2 = 1.3
    phi3 = -0.2
    p1 = QtCore.QPointF(x0 + r * cos(phi1), y0 + r * sin(phi1))
    p2 = QtCore.QPointF(x0 + r * cos(phi2), y0 + r * sin(phi2))
    p3 = QtCore.QPointF(x0 + r * cos(phi3), y0 + r * sin(phi3))

    n_angles = mask_model.find_n_angles_on_arc_from_three_points_around_p0(p0, p1, p2, p3, n)
    n_points = mask_model.calc_arc_points_from_angles(p0, r, width, n_angles)
    for p in n_points:
        rcalc = mask_model.find_radius_of_circle_from_center_and_point(p0, p)
        assert r == pytest.approx(rcalc, abs=1e-6)



import os
import pytest

import numpy as np
from xypattern import Pattern

from ...model.Configuration import Configuration
from ...model.BatchModel import BatchModel, iterate_folder

from mock import MagicMock

unittest_path = os.path.dirname(__file__)
data_path = os.path.join(unittest_path, "../data")
files = [
    os.path.join(data_path, "lambda/testasapo1_1009_00002_m1_part00000.nxs"),
    os.path.join(data_path, "lambda/testasapo1_1009_00002_m1_part00001.nxs"),
]

cal_file = os.path.join(data_path, "lambda/L2.poni")


@pytest.fixture()
def configuration():
    configuration = Configuration()
    yield configuration


@pytest.fixture()
def batch_model(configuration):
    configuration.calibration_model.load(cal_file)
    batch_model = BatchModel(configuration)
    batch_model.set_image_files(files)

    pattern = Pattern(
            x=np.linspace(0, 10, 100),
            y=np.random.random(100))

    configuration.calibration_model.integrate_1d = MagicMock(
        return_value=(pattern.x, pattern.y)
    )
    yield batch_model


def test_init_state(batch_model):
    assert batch_model.raw_available
    assert np.all(batch_model.file_map == [0, 10, 20])
    assert batch_model.n_img_all == 20
    assert batch_model.pos_map_all.shape == (20, 2)


def test_integrate_raw_data(batch_model):
    start = 2
    stop = 18
    step = 2

    batch_model.integrate_raw_data(start, stop, step, use_all=True)

    assert batch_model.data.shape[0] == 8
    assert batch_model.n_img == 8
    assert np.all(batch_model.pos_map[0] == [0, 2])
    assert batch_model.pos_map.shape == (8, 2)


def test_get_image_info(batch_model):
    image = 10
    name, pos = batch_model.get_image_info(image, use_all=True)
    assert name == files[1]
    assert pos == 0


def test_load_image(batch_model, configuration):
    index = 10
    batch_model.load_image(index, use_all=True)
    assert configuration.img_model.img_data.shape == (1833, 1556)


def test_saving_loading(batch_model, tmp_path):
    num_points = 1500
    start = 2
    stop = 18
    step = 2

    batch_model.integrate_raw_data(start, stop, step, use_all=True)
    batch_model.save_proc_data(os.path.join(tmp_path, "test_save_proc.nxs"))
    batch_model.reset_data()
    batch_model.load_proc_data(os.path.join(tmp_path, "test_save_proc.nxs"))

    assert batch_model.data.shape[0] == 8
    assert batch_model.n_img == 8
    assert np.all(batch_model.pos_map[0] == [0, 2])
    assert batch_model.pos_map.shape == (8, 2)


def test_save_as_csv(batch_model, tmp_path):
    batch_model.integrate_raw_data(start=5, stop=10, step=2, use_all=True)
    batch_model.save_as_csv(os.path.join(tmp_path, "test_save.csv"))
    assert os.path.exists(os.path.join(tmp_path, "test_save.csv"))


def test_extract_background(batch_model):
    batch_model.integrate_raw_data(start=5, stop=10, step=2, use_all=True)
    batch_model.extract_background(parameters=(0.1, 150, 50))
    assert batch_model.bkg.shape[0] == 3


def test_normalize(batch_model):
    batch_model.reset_data()
    batch_model.data = np.ones((3, 80))
    for i in range(batch_model.data.shape[0]):
        batch_model.data[i] *= np.random.random()

    batch_model.normalize()
    assert pytest.approx(0) == np.sum(np.diff(batch_model.data[:, 1]))


def test_iterate_folder():
    assert iterate_folder("r001", 1) == "r002"
    assert iterate_folder("r009", 1) == "r010"
    assert iterate_folder("r009", -1) == "r008"

    assert iterate_folder("test/r009", -1) == "test/r008"
    assert iterate_folder("test/r009", 1) == "test/r010"

    assert iterate_folder("exp/0250/test/r001", 1) == "exp/0250/test/r002"
    assert iterate_folder("exp/0250/test/r001", -1) == "exp/0250/test/r000"
    assert iterate_folder("exp/0250/test/r001", 10) == "exp/0250/test/r011"
    assert iterate_folder("exp/0250/test/r001", -10) == "exp/0250/test/r000"

    assert iterate_folder("exp234/02321/test/r099", 1) == "exp234/02321/test/r100"
    assert iterate_folder("exp234/02321/test/r099", -1) == "exp234/02321/test/r098"
    assert iterate_folder("exp234/02321/test/r099", 10) == "exp234/02321/test/r109"
    assert iterate_folder("exp234/02321/test/r099", -10) == "exp234/02321/test/r089"

    assert iterate_folder("exp234/02321/test/r999", 1) == "exp234/02321/test/r1000"
    assert iterate_folder("exp234/02321/test/r999", -1) == "exp234/02321/test/r998"
    assert iterate_folder("exp234/02321/test/r999", 10) == "exp234/02321/test/r1009"

    assert iterate_folder("exp234/02321/test/r100", -1) == "exp234/02321/test/r099"



# SPDX-License-Identifier: MIT

import os
import numpy as np
import json

from copy import deepcopy

from xypattern import Pattern
from xypattern.auto_background import SmoothBrucknerBackground

from .util import Signal
from .util.ImgCorrection import CbnCorrection, ObliqueAngleDetectorAbsorptionCorrection

from .util.calc import convert_units
from . import ImgModel, CalibrationModel, MaskModel, PatternModel, BatchModel
from .MapModel import MapModel
from .CalibrationModel import DetectorModes


def _json_numpy_default(obj):
    """JSON encoder default for numpy int/float types."""
    if isinstance(obj, np.integer):
        return int(obj)
    if isinstance(obj, np.floating):
        return float(obj)
    if isinstance(obj, np.ndarray):
        return obj.tolist()
    raise TypeError(f"Object of type {type(obj).__name__} is not JSON serializable")


class Configuration(object):
    """
    The configuration class contains a working combination of an ImgModel, PatternModel, MaskModel and CalibrationModel.
    It does handles the core data manipulation of Dioptas.
    The management of multiple Configurations is done by the DioptasModel.
    """

    def __init__(self, working_directories=None):
        super(Configuration, self).__init__()

        self.img_model = ImgModel()
        self.mask_model = MaskModel()
        self.calibration_model = CalibrationModel(self.img_model)
        self.pattern_model = PatternModel()

        self.batch_model = BatchModel(self)
        self.map_model = MapModel(self)

        if working_directories is None:
            self.working_directories = {
                "calibration": "",
                "mask": "",
                "image": os.path.expanduser("~"),
                "pattern": "",
                "overlay": "",
                "phase": "",
                "batch": os.path.expanduser("~"),
            }
        else:
            self.working_directories = working_directories

        self.use_mask = False

        self.transparent_mask = False

        self._integration_rad_points = None
        self._integration_unit = "2th_deg"
        self._oned_azimuth_range = None
        self.trim_trailing_zeros = True

        self._cake_azimuth_points = 360
        self._cake_azimuth_range = None

        self._auto_integrate_pattern = True
        self._auto_integrate_cake = False

        self.auto_save_integrated_pattern = False
        self.integrated_patterns_file_formats = [".xy"]

        self.cake_changed = Signal()
        self._connect_signals()

    def _connect_signals(self):
        """
        Connects the img_changed signal to responding functions.
        """
        self.img_model.img_changed.connect(self.update_mask_dimension)
        self.img_model.img_changed.connect(self.integrate_image_1d)

    def integrate_image_1d(self, update_pattern_model=True):
        """
        Integrates the image in the ImageModel to a Pattern. Will also automatically save the integrated pattern, if
        auto_save_integrated is True.

        :param update_pattern_model: If True, updates pattern_model and emits pattern_changed signal.
            Set to False during batch/map integration to avoid unnecessary GUI updates.
        """
        if self.calibration_model.is_calibrated:
            if self.use_mask:
                mask = self.mask_model.get_mask()
            elif self.mask_model.roi is not None:
                mask = self.mask_model.roi_mask
            else:
                mask = None

            x, y = self.calibration_model.integrate_1d(
                azi_range=self.oned_azimuth_range,
                mask=mask,
                unit=self.integration_unit,
                num_points=self.integration_rad_points,
                trim_zeros=self.trim_trailing_zeros,
            )

            if update_pattern_model:
                self.pattern_model.set_pattern(
                    x, y, self.img_model.filename, unit=self.integration_unit
                )

                if self.auto_save_integrated_pattern:
                    self._auto_save_patterns()

            return x, y

    def integrate_image_2d(self):
        """
        Integrates the image in the ImageModel to a Cake.
        """
        if self.use_mask:
            mask = self.mask_model.get_mask()
        elif self.mask_model.roi is not None:
            mask = self.mask_model.roi_mask
        else:
            mask = None

        self.calibration_model.integrate_2d(
            mask=mask,
            rad_points=self._integration_rad_points,
            azimuth_points=self._cake_azimuth_points,
            azimuth_range=self._cake_azimuth_range,
        )

        self.cake_changed.emit()

    def save_pattern(self, filename=None, subtract_background=False):
        """
        Saves the current integrated pattern. The format depends on the file ending. Possible file formats:
            [*.xy, *.chi, *.dat, *.fxye]
        :param filename: where to save the file
        :param subtract_background: flat whether the pattern should be saved with or without subtracted background
        """
        if filename is None:
            filename = self.img_model.filename

        if filename.endswith(".xy"):
            self.pattern_model.save_pattern(
                filename,
                header=self._create_xy_header(),
                subtract_background=subtract_background,
            )
        elif filename.endswith(".fxye"):
            self.pattern_model.save_pattern(
                filename,
                header=self._create_fxye_header(filename),
                subtract_background=subtract_background,
            )
        else:
            self.pattern_model.save_pattern(
                filename, subtract_background=subtract_background
            )

    def save_background_pattern(self, filename=None):
        """
        Saves the current fit background as a pattern. The format depends on the file ending. Possible file formats:
            [*.xy, *.chi, *.dat, *.fxye]
        """
        if filename is None:
            filename = self.img_model.filename

        if filename.endswith(".xy"):
            self.pattern_model.save_auto_background_as_pattern(
                filename, header=self._create_xy_header()
            )
        elif filename.endswith(".fxye"):
            self.pattern_model.save_auto_background_as_pattern(
                filename, header=self._create_fxye_header(filename)
            )
        else:
            self.pattern_model.save_pattern(filename)

    def _create_xy_header(self) -> str:
        """
        Creates the header for the xy file format (contains information about calibration parameters).
        :return: header string
        """
        header = self.calibration_model.create_file_header()
        header = header.replace("\r\n", "\n")
        header = header + "\n#\n# " + self._integration_unit + "\t I"
        return header

    def _create_fxye_header(self, filename) -> str:
        """
        Creates the header for the fxye file format (used by GSAS and GSAS-II) containing the calibration information
        :return: header string
        """
        header = "Generated file " + filename + " using DIOPTAS\n"
        header = header + self.calibration_model.create_file_header()
        unit = self._integration_unit
        lam = self.calibration_model.wavelength
        if unit == "q_A^-1":
            con = "CONQ"
        else:
            con = "CONS"

        header = (
            header
            + "\nBANK\t1\tNUM_POINTS\tNUM_POINTS "
            + con
            + "\tMIN_X_VAL\tSTEP_X_VAL "
            + "{0:.5g}".format(lam * 1e10)
            + " 0.0 FXYE"
        )
        return header

    def _auto_save_patterns(self):
        """
        Saves the current pattern in the pattern working directory (specified in self.working_directories['pattern'].
        When background subtraction is enabled in the pattern model the pattern will be saved with background
        subtraction and without in another sub-folder. ('bkg_subtracted')
        """
        for file_ending in self.integrated_patterns_file_formats:
            filename = os.path.join(
                self.working_directories["pattern"],
                os.path.basename(str(self.img_model.filename)).split(".")[:-1][0]
                + file_ending,
            )
            filename = filename.replace("\\", "/")
            self.save_pattern(filename, subtract_background=False)

        pattern = self.pattern_model.pattern

        if pattern.background_pattern is not None or pattern.auto_bkg is not None:
            for file_ending in self.integrated_patterns_file_formats:
                directory = os.path.join(
                    self.working_directories["pattern"], "bkg_subtracted"
                )
                if not os.path.exists(directory):
                    os.mkdir(directory)
                filename = os.path.join(
                    directory, self.pattern_model.pattern.name + file_ending
                )
                filename = filename.replace("\\", "/")
                self.save_pattern(filename, subtract_background=True)

    def update_mask_dimension(self):
        """
        Updates the shape of the mask in the MaskModel to the shape of the image in the ImageModel.
        """
        self.mask_model.set_dimension(self.img_model._img_data.shape)

    @property
    def integration_rad_points(self) -> int:
        return self._integration_rad_points

    @integration_rad_points.setter
    def integration_rad_points(self, new_value: int):
        self._integration_rad_points = new_value
        self.integrate_image_1d()
        if self.auto_integrate_cake:
            self.integrate_image_2d()

    @property
    def cake_azimuth_points(self) -> int:
        return self._cake_azimuth_points

    @cake_azimuth_points.setter
    def cake_azimuth_points(self, new_value):
        self._cake_azimuth_points = new_value
        if self.auto_integrate_cake:
            self.integrate_image_2d()

    @property
    def cake_azimuth_range(self):
        return self._cake_azimuth_range

    @cake_azimuth_range.setter
    def cake_azimuth_range(self, new_value):
        self._cake_azimuth_range = new_value
        if self.auto_integrate_cake:
            self.integrate_image_2d()

    @property
    def oned_azimuth_range(self):
        return self._oned_azimuth_range

    @oned_azimuth_range.setter
    def oned_azimuth_range(self, new_value):
        self._oned_azimuth_range = new_value
        if self.auto_integrate_pattern:
            self.integrate_image_1d()

    @property
    def integration_unit(self) -> str:
        return self._integration_unit

    @integration_unit.setter
    def integration_unit(self, new_unit: str):
        old_unit = self.integration_unit
        self._integration_unit = new_unit

        pattern = self.pattern_model.pattern
        x = getattr(pattern, "x", None)
        valid_units = {"2th_deg", "q_A^-1", "d_A"}
        if old_unit not in valid_units or new_unit not in valid_units:
            return
        if old_unit == new_unit:
            return
        if x is not None and len(x) > 1:
            pattern.transform_x(
                lambda x: convert_units(
                    x, self.calibration_model.wavelength, old_unit, new_unit
                )
            )
            self.integrate_image_1d()

    @property
    def correct_solid_angle(self):
        return self.calibration_model.correct_solid_angle

    @correct_solid_angle.setter
    def correct_solid_angle(self, new_val):
        self.calibration_model.correct_solid_angle = new_val
        if self.auto_integrate_pattern:
            self.integrate_image_1d()
        if self._auto_integrate_cake:
            self.integrate_image_2d()

    @property
    def is_calibrated(self) -> bool:
        return self.calibration_model.is_calibrated

    @property
    def auto_integrate_cake(self) -> bool:
        return self._auto_integrate_cake

    @auto_integrate_cake.setter
    def auto_integrate_cake(self, new_value):
        if self._auto_integrate_cake == new_value:
            return

        self._auto_integrate_cake = new_value
        if new_value:
            self.img_model.img_changed.connect(self.integrate_image_2d)
        else:
            self.img_model.img_changed.disconnect(self.integrate_image_2d)

    @property
    def auto_integrate_pattern(self) -> bool:
        return self._auto_integrate_pattern

    @auto_integrate_pattern.setter
    def auto_integrate_pattern(self, new_value):
        if self._auto_integrate_pattern == new_value:
            return

        self._auto_integrate_pattern = new_value
        if new_value:
            self.img_model.img_changed.connect(self.integrate_image_1d)
        else:
            self.img_model.img_changed.disconnect(self.integrate_image_1d)

    @property
    def cake_img(self) -> np.ndarray:
        return self.calibration_model.cake_img

    @property
    def roi(self):
        return self.mask_model.roi

    @roi.setter
    def roi(self, new_val):
        self.mask_model.roi = new_val
        self.integrate_image_1d()

    def copy(self) -> "Configuration":
        """
        Creates a copy of the current configuration directory
        :return: copied configuration
        :rtype: Configuration
        """
        new_configuration = Configuration(self.working_directories)
        new_configuration.img_model._img_data = self.img_model._img_data
        new_configuration.img_model.img_transformations = deepcopy(
            self.img_model.img_transformations
        )

        new_configuration.calibration_model.set_pyFAI(
            self.calibration_model.get_calibration_parameter()[0]
        )
        new_configuration.integrate_image_1d()

        return new_configuration

    def save_in_hdf5(self, hdf5_group):
        """
        Saves the configuration group in the given hdf5_group.
        :type hdf5_group: h5py.Group
        """

        f = hdf5_group
        # save general information
        general_information = f.create_group("general_information")
        # integration parameters:
        general_information.attrs["integration_unit"] = self.integration_unit
        if self.integration_rad_points:
            general_information.attrs["integration_num_points"] = (
                self.integration_rad_points
            )
        else:
            general_information.attrs["integration_num_points"] = 0

        # cake parameters:
        general_information.attrs["auto_integrate_cake"] = self.auto_integrate_cake
        general_information.attrs["cake_azimuth_points"] = self.cake_azimuth_points
        if self.cake_azimuth_range is None:
            general_information.attrs["cake_azimuth_range"] = "None"
        else:
            general_information.attrs["cake_azimuth_range"] = self.cake_azimuth_range

        # mask parameters
        general_information.attrs["use_mask"] = self.use_mask
        general_information.attrs["transparent_mask"] = self.transparent_mask

        # auto save parameters
        general_information.attrs["auto_save_integrated_pattern"] = (
            self.auto_save_integrated_pattern
        )
        formats = [
            n.encode("ascii", "ignore") for n in self.integrated_patterns_file_formats
        ]
        general_information.create_dataset(
            "integrated_patterns_file_formats", (len(formats), 1), "S10", formats
        )

        # save working directories
        working_directories_gp = f.create_group("working_directories")
        try:
            for key in self.working_directories:
                working_directories_gp.attrs[key] = self.working_directories[key]
        except TypeError:
            self.working_directories = {
                "calibration": "",
                "mask": "",
                "image": "",
                "pattern": "",
                "overlay": "",
                "phase": "",
                "batch": "",
            }
            for key in self.working_directories:
                working_directories_gp.attrs[key] = self.working_directories[key]

        # save image model
        image_group = f.create_group("image_model")
        image_group.attrs["auto_process"] = self.img_model.autoprocess
        image_group.attrs["factor"] = self.img_model.factor
        image_group.attrs["has_background"] = self.img_model.has_background()
        image_group.attrs["background_filename"] = self.img_model.background_filename
        image_group.attrs["background_offset"] = self.img_model.background_offset
        image_group.attrs["background_scaling"] = self.img_model.background_scaling
        if self.img_model.has_background():
            background_data = self.img_model.untransformed_background_data
            image_group.create_dataset(
                "background_data", background_data.shape, "f", background_data
            )

        image_group.attrs["series_max"] = self.img_model.series_max
        image_group.attrs["series_pos"] = self.img_model.series_pos

        # image corrections
        corrections_group = image_group.create_group("corrections")
        corrections_group.attrs["has_corrections"] = self.img_model.has_corrections()
        for (
            correction,
            correction_object,
        ) in self.img_model.img_corrections.corrections.items():
            if correction in ["cbn", "oiadac"]:
                correction_data = correction_object.get_data()
                imcd = corrections_group.create_dataset(
                    correction, correction_data.shape, "f", correction_data
                )
                for param, value in correction_object.get_params().items():
                    imcd.attrs[param] = value
            elif correction == "transfer":
                params = correction_object.get_params()
                transfer_group = corrections_group.create_group("transfer")
                original_data = params["original_data"]
                response_data = params["response_data"]
                original_ds = transfer_group.create_dataset(
                    "original_data", original_data.shape, "f", original_data
                )
                original_ds.attrs["filename"] = params["original_filename"]
                response_ds = transfer_group.create_dataset(
                    "response_data", response_data.shape, "f", response_data
                )
                response_ds.attrs["filename"] = params["response_filename"]

        # the actual image
        image_group.attrs["filename"] = self.img_model.filename
        current_raw_image = self.img_model.untransformed_raw_img_data

        raw_image_data = image_group.create_dataset(
            "raw_image_data", current_raw_image.shape, dtype="f"
        )
        raw_image_data[...] = current_raw_image

        # image transformations
        transformations_group = image_group.create_group("image_transformations")
        for ind, transformation in enumerate(
            self.img_model.get_transformations_string_list()
        ):
            transformations_group.attrs[str(ind)] = transformation

        # save roi data
        if self.roi is not None:
            image_group.attrs["has_roi"] = True
            image_group.create_dataset("roi", (4,), "i8", tuple(self.roi))
        else:
            image_group.attrs["has_roi"] = False

        # save mask model
        mask_group = f.create_group("mask")
        current_mask = self.mask_model.get_mask()
        mask_data = mask_group.create_dataset("data", current_mask.shape, dtype=bool)
        mask_data[...] = current_mask

        # save detector information
        detector_group = f.create_group("detector")
        detector_mode = self.calibration_model.detector_mode
        detector_group.attrs["detector_mode"] = detector_mode.value
        if detector_mode == DetectorModes.PREDEFINED:
            detector_group.attrs["detector_name"] = self.calibration_model.detector.name
        elif detector_mode == DetectorModes.NEXUS:
            detector_group.attrs["nexus_filename"] = (
                self.calibration_model.detector.filename
            )

        # save calibration model
        calibration_group = f.create_group("calibration_model")
        # version 2.0 is used to indicate that the pyFAI parameters are stored as a json string
        calibration_group.attrs["version"] = "2.0"

        calibration_filename = self.calibration_model.filename
        if calibration_filename.endswith(".poni"):
            base_filename, ext = self.calibration_model.filename.rsplit(".", 1)
        else:
            base_filename = self.calibration_model.filename
            ext = "poni"
        calibration_group.attrs["calibration_filename"] = base_filename + "." + ext

        pyfai_config = self.calibration_model.pattern_geometry.get_config()
        calibration_group.attrs["pyfai_parameters"] = json.dumps(
            pyfai_config, default=_json_numpy_default
        )
        calibration_group.attrs["polarization_factor"] = (
            self.calibration_model.polarization_factor
        )

        calibration_group.attrs["correct_solid_angle"] = self.correct_solid_angle
        if self.calibration_model.distortion_spline_filename is not None:
            calibration_group.attrs["distortion_spline_filename"] = (
                self.calibration_model.distortion_spline_filename
            )

        # save background pattern and pattern model
        background_pattern_group = f.create_group("background_pattern")
        try:
            background_pattern_x = self.pattern_model.background_pattern._original_x
            background_pattern_y = self.pattern_model.background_pattern._original_y
        except (TypeError, AttributeError):
            background_pattern_x = None
            background_pattern_y = None
        if background_pattern_x is not None and background_pattern_y is not None:
            background_pattern_group.attrs["has_background_pattern"] = True
            bgx = background_pattern_group.create_dataset(
                "x", background_pattern_x.shape, dtype="f"
            )
            bgy = background_pattern_group.create_dataset(
                "y", background_pattern_y.shape, dtype="f"
            )
            bgx[...] = background_pattern_x
            bgy[...] = background_pattern_y
        else:
            background_pattern_group.attrs["has_background_pattern"] = False

        pattern_group = f.create_group("pattern")
        try:
            pattern_x = self.pattern_model.pattern._original_x
            pattern_y = self.pattern_model.pattern._original_y
        except (TypeError, AttributeError):
            pattern_x = None
            pattern_y = None
        if pattern_x is not None and pattern_y is not None:
            px = pattern_group.create_dataset("x", pattern_x.shape, dtype="f")
            py = pattern_group.create_dataset("y", pattern_y.shape, dtype="f")
            px[...] = pattern_x
            py[...] = pattern_y
        pattern_group.attrs["pattern_filename"] = self.pattern_model.pattern_filename
        pattern_group.attrs["unit"] = self.pattern_model.unit
        pattern_group.attrs["file_iteration_mode"] = (
            self.pattern_model.file_iteration_mode
        )
        if self.pattern_model.pattern.auto_bkg:
            pattern_group.attrs["auto_background_subtraction"] = True
            auto_background_group = pattern_group.create_group(
                "auto_background_settings"
            )
            auto_bkg = self.pattern_model.pattern.auto_bkg

            if type(auto_bkg) == SmoothBrucknerBackground:
                auto_background_group.attrs["smoothing"] = auto_bkg.smooth_width
                auto_background_group.attrs["iterations"] = auto_bkg.iterations
                auto_background_group.attrs["poly_order"] = auto_bkg.cheb_order

            auto_bkg_roi = self.pattern_model.pattern.auto_bkg_roi
            if auto_bkg_roi is not None:
                auto_background_group.attrs["x_start"] = auto_bkg_roi[0]
                auto_background_group.attrs["x_end"] = auto_bkg_roi[1]
        else:
            pattern_group.attrs["auto_background_subtraction"] = False

        # save map model
        self.map_model.save_in_hdf5(f)

    def load_from_hdf5(self, hdf5_group):
        """
        Loads a configuration from the specified hdf5_group.
        :type hdf5_group: h5py.Group
        """

        f = hdf5_group

        # disable all automatic functions
        self.auto_integrate_pattern = False
        self.auto_integrate_cake = False
        self.auto_save_integrated_pattern = False

        # get working directories
        working_directories = {}
        for key, value in f.get("working_directories").attrs.items():
            if os.path.isdir(value):
                working_directories[key] = value
            else:
                working_directories[key] = ""
        self.working_directories = working_directories

        # load pyFAI parameters
        try:
            calibration_schema_version = f.get("calibration_model").attrs["version"]
            if calibration_schema_version == "2.0":
                # pyFAI parameters are stored as a json string
                pyfai_config = json.loads(
                    f.get("calibration_model").attrs["pyfai_parameters"]
                )
                self.calibration_model.set_pyFAI_config(pyfai_config)

                # polarization factor is stored as an attribute (for some reason not in the pyFAI config)
                self.calibration_model.polarization_factor = f.get(
                    "calibration_model"
                ).attrs["polarization_factor"]
        except (KeyError, ValueError):
            # if the version is not set, we assume that the pyFAI parameters are stored in the old way
            # pyFAI parameters are stored as attributes of the group
            # this is the old way of storing pyFAI parameters
            # and is kept for backwards compatibility
            pyfai_parameters = {}
            pyfai_parameters_group = f.get("calibration_model").get("pyfai_parameters")
            if pyfai_parameters_group is not None:
                for key, value in pyfai_parameters_group.attrs.items():
                    pyfai_parameters[key] = value

            try:
                if pyfai_parameters:
                    self.calibration_model.set_pyFAI(pyfai_parameters)

            except (KeyError, ValueError):
                print("Problem with saved pyFAI calibration parameters")
                pass

        filename = f.get("calibration_model").attrs["calibration_filename"]
        (_, base_name) = os.path.split(filename)
        self.calibration_model.filename = filename
        self.calibration_model.calibration_name = base_name

        try:
            self.correct_solid_angle = f.get("calibration_model").attrs[
                "correct_solid_angle"
            ]
        except KeyError:
            pass

        try:
            distortion_spline_filename = f.get("calibration_model").attrs[
                "distortion_spline_filename"
            ]
            self.calibration_model.load_distortion(distortion_spline_filename)
        except KeyError:
            pass

        # load detector definition
        try:
            detector_mode = f.get("detector").attrs["detector_mode"]
            if detector_mode == DetectorModes.PREDEFINED.value:
                detector_name = f.get("detector").attrs["detector_name"]
                self.calibration_model.load_detector(detector_name)
            elif detector_mode == DetectorModes.NEXUS.value:
                nexus_filename = f.get("detector").attrs["nexus_filename"]
                self.calibration_model.load_detector_from_file(nexus_filename)
        except AttributeError:  # to ensure backwards compatibility
            pass

        # load img_model
        self.img_model._img_data = np.copy(
            f.get("image_model").get("raw_image_data")[...]
        )
        filename = f.get("image_model").attrs["filename"]
        self.img_model.filename = filename

        try:
            self.img_model.file_name_iterator.update_filename(filename)
            self.img_model._directory_watcher.path = os.path.dirname(filename)
        except EnvironmentError:
            pass

        self.img_model.autoprocess = f.get("image_model").attrs["auto_process"]
        self.img_model.autoprocess_changed.emit()
        self.img_model.factor = f.get("image_model").attrs["factor"]

        try:
            self.img_model.series_max = f.get("image_model").attrs["series_max"]
            self.img_model.series_pos = f.get("image_model").attrs["series_pos"]
        except KeyError:
            pass

        if f.get("image_model").attrs["has_background"]:
            self.img_model.background_data = np.copy(
                f.get("image_model").get("background_data")[...]
            )
            self.img_model.background_filename = f.get("image_model").attrs[
                "background_filename"
            ]
            self.img_model.background_scaling = f.get("image_model").attrs[
                "background_scaling"
            ]
            self.img_model.background_offset = f.get("image_model").attrs[
                "background_offset"
            ]

        # load image transformations
        transformation_group = f.get("image_model").get("image_transformations")
        transformation_list = []
        for key, transformation in transformation_group.attrs.items():
            transformation_list.append(transformation)
        self.calibration_model.load_transformations_string_list(transformation_list)
        self.img_model.load_transformations_string_list(transformation_list)

        # load roi data
        if f.get("image_model").attrs["has_roi"]:
            self.roi = tuple(f.get("image_model").get("roi")[...])

        # load mask model
        self.mask_model.set_mask(np.copy(f.get("mask").get("data")[...]))

        # load pattern model
        if f.get("pattern").get("x") and f.get("pattern").get("y"):
            self.pattern_model.set_pattern(
                f.get("pattern").get("x")[...],
                f.get("pattern").get("y")[...],
                f.get("pattern").attrs["pattern_filename"],
                f.get("pattern").attrs["unit"],
            )
            self.pattern_model.file_iteration_mode = f.get("pattern").attrs[
                "file_iteration_mode"
            ]
        self.integration_unit = f.get("general_information").attrs["integration_unit"]

        if f.get("background_pattern").attrs["has_background_pattern"]:
            self.pattern_model.background_pattern = Pattern(
                f.get("background_pattern").get("x")[...],
                f.get("background_pattern").get("y")[...],
                "background_pattern",
            )

        if f.get("pattern").attrs["auto_background_subtraction"]:
            self.pattern_model.pattern.auto_bkg_roi = [
                f.get("pattern").get("auto_background_settings").attrs["x_start"],
                f.get("pattern").get("auto_background_settings").attrs["x_end"],
            ]
            self.pattern_model.pattern.auto_bkg = SmoothBrucknerBackground(
                f.get("pattern").get("auto_background_settings").attrs["smoothing"],
                f.get("pattern").get("auto_background_settings").attrs["iterations"],
                f.get("pattern").get("auto_background_settings").attrs["poly_order"],
            )

        # load general configuration
        if f.get("general_information").attrs["integration_num_points"]:
            self.integration_rad_points = int(
                f.get("general_information").attrs["integration_num_points"]
            )

        # cake parameters:
        self.auto_integrate_cake = f.get("general_information").attrs[
            "auto_integrate_cake"
        ]
        try:
            self.cake_azimuth_points = f.get("general_information").attrs[
                "cake_azimuth_points"
            ]
        except KeyError as e:
            pass
        try:
            if f.get("general_information").attrs["cake_azimuth_range"] == "None":
                self.cake_azimuth_range = None
            else:
                self.cake_azimuth_range = f.get("general_information").attrs[
                    "cake_azimuth_range"
                ]
        except KeyError as e:
            pass

        # mask parameters
        self.use_mask = f.get("general_information").attrs["use_mask"]
        self.transparent_mask = f.get("general_information").attrs["transparent_mask"]

        # corrections
        if f.get("image_model").get("corrections").attrs["has_corrections"]:
            for name, correction_group in (
                f.get("image_model").get("corrections").items()
            ):
                params = {}
                for param, val in correction_group.attrs.items():
                    params[param] = val
                if name == "cbn":
                    tth_array = (
                        180.0 / np.pi * self.calibration_model.tth_array
                    )
                    azi_array = (
                        180.0 / np.pi * self.calibration_model.azi_array
                    )
                    cbn_correction = CbnCorrection(
                        tth_array=tth_array, azi_array=azi_array
                    )

                    cbn_correction.set_params(params)
                    cbn_correction.update()
                    self.img_model.add_img_correction(cbn_correction, name)
                elif name == "oiadac":
                    tth_array = (
                        180.0 / np.pi * self.calibration_model.tth_array
                    )
                    azi_array = (
                        180.0 / np.pi * self.calibration_model.azi_array
                    )
                    oiadac = ObliqueAngleDetectorAbsorptionCorrection(
                        tth_array=tth_array, azi_array=azi_array
                    )

                    oiadac.set_params(params)
                    oiadac.update()
                    self.img_model.add_img_correction(oiadac, name)
                elif name == "transfer":
                    params = {
                        "original_data": correction_group.get("original_data")[...],
                        "original_filename": correction_group.get(
                            "original_data"
                        ).attrs["filename"],
                        "response_data": correction_group.get("response_data")[...],
                        "response_filename": correction_group.get(
                            "response_data"
                        ).attrs["filename"],
                    }

                    self.img_model.transfer_correction.set_params(params)
                    self.img_model.enable_transfer_function()

        # autosave parameters
        self.auto_save_integrated_pattern = f.get("general_information").attrs[
            "auto_save_integrated_pattern"
        ]
        self.integrated_patterns_file_formats = []
        for file_format in f.get("general_information").get(
            "integrated_patterns_file_formats"
        ):
            self.integrated_patterns_file_formats.append(file_format[0].decode("utf-8"))

        if self.calibration_model.is_calibrated:
            self.integrate_image_1d()
        else:
            self.pattern_model.pattern.recalculate_pattern()

        # load map model
        self.map_model.load_from_hdf5(f)



# SPDX-License-Identifier: MIT

import os
import sys
from sys import platform as _platform

# If QT_API is not set, use PyQt6 by default
if "QT_API" not in os.environ:
    try:
        import PyQt6.QtCore
    except ImportError:
        pass

from qtpy import QtWidgets
from qt_material import apply_stylesheet

try:
    from pyshortcuts import make_shortcut
except ImportError:
    make_shortcut = None

try:
    from ._version import version as __version__
except Exception:
    try:
        from importlib.metadata import version as _pkg_version
        __version__ = _pkg_version("dioptas")
    except Exception:
        __version__ = "0.0.0"

from .paths import resources_path, calibrants_path, icons_path, data_path, style_path
from .excepthook import excepthook
from .controller.MainController import MainController


theme_path = os.path.join(style_path, "dark_orange.xml")
qss_path = os.path.join(style_path, "qt_material.css")

_dioptrin_available = False


def _check_dioptrin_license():
    """Check dioptrin license at startup. Returns True if usable."""
    try:
        import dioptrin

        dioptrin.validate_license()
        return True
    except ImportError:
        return False
    except dioptrin.LicenseNotFoundError:
        return False
    except dioptrin.LicenseExpiredError:
        QtWidgets.QMessageBox.warning(
            None,
            "Dioptrin License Expired",
            "Your Dioptrin license has expired. "
            "Dioptas will use pyFAI for integration.\n\n"
            "Please renew your license to continue using Dioptrin.",
        )
        return False
    except dioptrin.LicenseError:
        return False


def main():
    global _dioptrin_available

    app = QtWidgets.QApplication([])

    apply_stylesheet(
        app,
        theme=theme_path,
        css_file=qss_path,
        extra={"density_scale": -2},
    )
    sys.excepthook = excepthook
    print("Dioptas {}".format(__version__))

    _dioptrin_available = _check_dioptrin_license()

    if len(sys.argv) == 1:  # normal start
        controller = MainController()
        controller.show_window()
        app.exec_()
    else:  # with command line arguments
        if sys.argv[1] == "test":
            controller = MainController(use_settings=False)
            controller.show_window()

        elif sys.argv[1].startswith("makeshortcut"):
            if make_shortcut is None:
                raise ImportError("pyshortcuts not installed.  Try `pip install pyshortcuts`")
            make_shortcut(
                "-m dioptas",
                name="Dioptas",
                description="Dioptas 2D XRD {}".format(__version__),
                icon=os.path.join(icons_path, "icon"),
                terminal=False,
            )

        elif sys.argv[1].startswith("version"):
            print(__version__)

        elif sys.argv[1].endswith(".json"):
            controller = MainController(config_file=sys.argv[1])
            controller.show_window()
            app.exec_()
    del app



# SPDX-License-Identifier: MIT

import os
import weakref
import pytest


@pytest.fixture(scope="session")
def qapp():
    """Fixture ensuring QApplication is instanciated"""
    from qtpy import QtWidgets

    app = QtWidgets.QApplication.instance()
    if app is None:
        app = QtWidgets.QApplication([])
    try:
        yield app
    finally:
        if app is not None:
            app.closeAllWindows()


@pytest.fixture
def qWidgetFactory(qapp):
    """QWidget factory as fixture

    This fixture provides a function taking a QWidget subclass as argument
    which returns an instance of this QWidget making sure it is shown first
    and destroyed once the test is done.
    """
    from qtpy import QtCore
    from qtpy.QtTest import QTest

    widgets = set()

    def createWidget(cls, *args, **kwargs):
        widget = cls(*args, **kwargs)
        # For Popup windows, remove the Popup flag to prevent automatic closing in tests
        if widget.windowFlags() & QtCore.Qt.Popup:
            widget.setWindowFlags(widget.windowFlags() & ~QtCore.Qt.Popup)
        # Don't set WA_DeleteOnClose if widget already has it
        if not widget.testAttribute(QtCore.Qt.WA_DeleteOnClose):
            widget.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        widget.show()
        QTest.qWaitForWindowExposed(widget)
        widgets.add(widget)
        return weakref.proxy(widget)

    try:
        yield createWidget
    finally:
        for widget in widgets:
            widget.close()
        qapp.processEvents()


@pytest.fixture
def main_controller(qapp):
    """Fixture providing a MainController instance"""
    from qtpy.QtTest import QTest
    from dioptas.controller.MainController import MainController
    from qtpy import QtCore

    controller = MainController(use_settings=False)
    controller.show_window()
    controller.widget.setAttribute(QtCore.Qt.WA_DeleteOnClose)
    QTest.qWaitForWindowExposed(controller.widget)
    try:
        yield controller
    finally:
        controller.widget.close()


@pytest.fixture(scope="function")
def dioptas_model():
    from dioptas.model.DioptasModel import DioptasModel

    model = DioptasModel()
    yield model


@pytest.fixture
def phase_controller(integration_widget, dioptas_model):
    from dioptas.controller.integration import PhaseController

    return PhaseController(integration_widget, dioptas_model)


@pytest.fixture
def pattern_controller(integration_widget, dioptas_model):
    from dioptas.controller.integration import PatternController

    return PatternController(integration_widget, dioptas_model)


@pytest.fixture
def integration_widget(qtbot):
    from dioptas.widgets.integration import IntegrationWidget

    widget = IntegrationWidget()
    yield widget
    widget.close()


@pytest.fixture
def calibration_widget(qtbot):
    from dioptas.widgets.CalibrationWidget import CalibrationWidget

    widget = CalibrationWidget()
    yield widget
    widget.close()


@pytest.fixture
def integration_controller(integration_widget, dioptas_model, qtbot):
    from dioptas.controller.integration import IntegrationController

    return IntegrationController(widget=integration_widget, dioptas_model=dioptas_model)


@pytest.fixture
def batch_model(dioptas_model):
    return dioptas_model.batch_model


@pytest.fixture
def batch_controller(integration_widget, dioptas_model):
    from dioptas.controller.integration import BatchController

    return BatchController(integration_widget, dioptas_model)


@pytest.fixture
def batch_widget(integration_widget):
    return integration_widget.batch_widget


@pytest.fixture
def background_controller(integration_widget, dioptas_model, qtbot):
    from dioptas.controller.integration import BackgroundController
    return BackgroundController(integration_widget, dioptas_model)


@pytest.fixture
def image_controller(integration_widget, dioptas_model, qtbot):
    from dioptas.controller.integration import ImageController
    return ImageController(integration_widget, dioptas_model)


@pytest.fixture
def calibration_controller(calibration_widget, dioptas_model, qtbot):
    from dioptas.controller import CalibrationController
    return CalibrationController(calibration_widget, dioptas_model)


@pytest.fixture
def calibration_model(dioptas_model):
    return dioptas_model.calibration_model


@pytest.fixture
def img_model(dioptas_model):
    return dioptas_model.img_model



import os
import numpy as np
import pytest
from pytest import approx
from qtpy import QtWidgets, QtCore
from qtpy.QtTest import QTest

from mock import MagicMock
from dioptas.controller.MapController import MapController
from dioptas.model.DioptasModel import DioptasModel

from dioptas.model.MapModel import MapModel, MapPointInfo, create_map
from dioptas.widgets.MapWidget import MapWidget
from dioptas.widgets.plot_widgets.PatternWidget import SymmetricModifiedLinearRegionItem


unittest_path = os.path.dirname(__file__)
data_path = os.path.join(unittest_path, os.pardir, "data")
map_img_path = os.path.join(data_path, "map")
map_pattern_path = os.path.join(data_path, "map", "xy")
map_img_file_names = [
    f for f in os.listdir(map_img_path) if os.path.isfile(os.path.join(map_img_path, f))
]
map_img_file_paths = [
    os.path.join(map_img_path, filename) for filename in map_img_file_names
]


@pytest.fixture
def dioptas_model():
    return DioptasModel()


@pytest.fixture
def map_controller(qapp, dioptas_model: DioptasModel):
    """Fixture providing a MainController instance"""
    widget = MapWidget()
    model = dioptas_model
    controller = MapController(widget, model)
    controller.widget.setAttribute(QtCore.Qt.WA_DeleteOnClose)
    QTest.qWaitForWindowExposed(controller.widget)
    try:
        yield controller
    finally:
        controller.widget.close()


@pytest.fixture
def map_model(map_controller) -> MapModel:
    return map_controller.model.map_model


def load_calibration(map_controller: MapController):
    map_controller.model.current_configuration.calibration_model.load(
        os.path.join(data_path, "CeO2_Pilatus1M.poni")
    )


def mock_open_filenames(filepaths):
    QtWidgets.QFileDialog.getOpenFileNames = MagicMock(return_value=filepaths)


def mock_save_filename(filepath):
    QtWidgets.QFileDialog.getSaveFileName = MagicMock(return_value=filepath)


def mock_map_model(map_model: MapModel):
    map_model.map = create_map(np.array([1, 2, 3, 4, 5, 6]), (2, 3))
    map_model.filepaths = map_img_file_paths
    map_model.possible_dimensions = [(1, 6), (2, 3), (3, 2), (6, 1)]
    map_model.point_infos = [MapPointInfo(f) for f in map_img_file_paths]
    map_model.dimension = (2, 3)
    map_model.map_changed.emit()


def mock_integrate_1d(map_controller: MapController):
    map_controller.model.calibration_model.integrate_1d = MagicMock(
        return_value=(np.arange(10), np.arange(10))
    )


def test_click_load_starts_creating_map(map_controller, map_model: MapModel):
    map_model.load = MagicMock()
    mock_open_filenames(map_img_file_paths)
    mock_integrate_1d(map_controller)
    map_controller.load_btn_clicked()
    map_model.load.assert_called_once()
    assert map_model.load.call_args[0][0] == map_img_file_paths


def test_click_load_empties_file_list_without_calibration(
    map_controller, map_model: MapModel
):
    mock_open_filenames(map_img_file_paths)
    QtWidgets.QMessageBox.critical = MagicMock()
    map_controller.load_btn_clicked()

    assert map_controller.widget.control_widget.file_list.count() == 0
    assert QtWidgets.QMessageBox.critical.assert_called_once


def test_load_empty_filelist(map_controller, map_model: MapModel):
    mock_open_filenames([])
    map_controller.load_btn_clicked()

    assert map_controller.widget.control_widget.file_list.count() == 0


def test_files_with_different_dimensions(map_controller, map_model: MapModel):
    load_calibration(map_controller)
    mock_open_filenames(
        [
            os.path.join(data_path, "CeO2_Pilatus1M.tif"),
            os.path.join(data_path, "image_001.tif"),
        ]
    )
    QtWidgets.QMessageBox.critical = MagicMock()
    map_controller.load_btn_clicked()
    assert QtWidgets.QMessageBox.critical.assert_called_once
    assert map_model.filepaths is None
    assert map_controller.widget.control_widget.file_list.count() == 0


def test_files_with_different_dimensions_shows_error_dialog(
    map_controller, map_model: MapModel
):
    """RuntimeError from dioptrin shape mismatch is caught and shown as error dialog."""
    mock_open_filenames(map_img_file_paths)
    map_model.load = MagicMock(
        side_effect=RuntimeError("mask dim1 mismatch: mask = 2048, data = 3262")
    )
    QtWidgets.QMessageBox.critical = MagicMock()
    map_controller.load_btn_clicked()
    QtWidgets.QMessageBox.critical.assert_called_once()
    assert "mask dim1 mismatch" in str(
        QtWidgets.QMessageBox.critical.call_args
    )


def test_click_load_fills_file_list(map_controller, map_model: MapModel):
    load_calibration(map_controller)
    assert map_controller.model.current_configuration.is_calibrated == True
    mock_open_filenames(map_img_file_paths)
    mock_integrate_1d(map_controller)
    map_controller.load_btn_clicked()

    assert map_model.filepaths == map_img_file_paths
    assert map_controller.widget.control_widget.file_list.count() == len(
        map_img_file_paths
    )


def test_mask_is_shown(map_controller):
    img_model = map_controller.model.img_model
    mask_model = map_controller.model.mask_model
    img_model.load(map_img_file_paths[0])

    map_controller.model.use_mask = True

    mask = np.zeros_like(img_model.img_data, dtype=bool)
    mask[0, 0] = True
    mask_model.set_mask(mask)
    img_model.img_changed.emit()

    assert map_controller.widget.img_plot_widget.mask_data is not None
    assert np.array_equal(map_controller.widget.img_plot_widget.mask_data, mask)


def test_loading_files_plots_map(map_controller: MapController, map_model: MapModel):
    load_calibration(map_controller)
    assert map_controller.model.current_configuration.is_calibrated == True
    assert map_controller.widget.map_plot_widget.img_data is None

    mock_open_filenames(map_img_file_paths)
    mock_integrate_1d(map_controller)
    map_controller.load_btn_clicked()

    plot_widget = map_controller.widget.map_plot_widget

    assert map_model.map is not None
    assert plot_widget.img_data is not None
    assert plot_widget.img_data.shape == map_model.map.shape
    assert plot_widget.data_img_item.image is not None


def test_loading_files_also_plots_first_image(
    map_controller: MapController, map_model: MapModel
):
    load_calibration(map_controller)
    assert map_controller.model.current_configuration.is_calibrated == True
    mock_open_filenames(map_img_file_paths)
    mock_integrate_1d(map_controller)
    map_controller.load_btn_clicked()

    plot_widget = map_controller.widget.img_plot_widget

    assert plot_widget.img_data is not None
    assert (
        plot_widget.img_data.shape == map_model.configuration.img_model.img_data.shape
    )
    assert np.array_equal(
        plot_widget.img_data, map_model.configuration.img_model.img_data
    )
    assert plot_widget.data_img_item.image is not None


def test_click_load_shows_error_if_not_calibrated(map_controller):
    mock_open_filenames(map_img_file_paths)
    assert map_controller.model.current_configuration.is_calibrated == False
    QtWidgets.QMessageBox.critical = MagicMock()
    map_controller.load_btn_clicked()
    assert QtWidgets.QMessageBox.critical.assert_called_once


def test_select_file_in_file_list_will_update_gui(map_controller):
    load_calibration(map_controller)
    mock_open_filenames(map_img_file_paths)
    map_controller.load_btn_clicked()

    # cache current image
    current_img = map_controller.widget.img_plot_widget.img_data.copy()

    # select second file in file list
    map_controller.widget.control_widget.file_list.setCurrentRow(1)
    assert (
        map_controller.model.current_configuration.img_model.filename
        == map_img_file_paths[1]
    )

    # check that image has changed
    assert not np.array_equal(
        map_controller.widget.img_plot_widget.img_data, current_img
    )


def test_mouse_click_item_in_map_plot_widget_updates_correctly(
    map_controller, dioptas_model
):
    load_calibration(map_controller)
    mock_open_filenames(map_img_file_paths)
    map_controller.load_btn_clicked()

    click_x, click_y = map_controller.widget.map_plot_widget.mouse_click_item.getData()
    assert click_x[0] == approx(0)
    assert click_y[0] == approx(0)

    map_controller.widget.control_widget.file_list.setCurrentRow(1)
    # check that mouse click item in map_plot_widget has changed
    click_x, click_y = map_controller.widget.map_plot_widget.mouse_click_item.getData()
    assert click_x[0] == approx(1.5)
    assert click_y[0] == approx(2.5)

    # check that replotting does not change po
    map_controller.widget.pattern_plot_widget.mouse_left_clicked.emit(10, 0)

    click_x, click_y = map_controller.widget.map_plot_widget.mouse_click_item.getData()
    assert click_x[0] == approx(1.5)
    assert click_y[0] == approx(2.5)


def test_select_file_in_file_list_integrates_1d_only_once(map_controller):
    load_calibration(map_controller)
    mock_open_filenames(map_img_file_paths[:2])
    mock_integrate_1d(map_controller)
    map_controller.load_btn_clicked()
    load_call_count = map_controller.model.calibration_model.integrate_1d.call_count
    map_controller.widget.control_widget.file_list.setCurrentRow(1)
    assert (
        map_controller.model.calibration_model.integrate_1d.call_count
        == load_call_count + 1
    )


def test_click_in_map_image_will_update_gui(map_controller):
    load_calibration(map_controller)
    mock_open_filenames(map_img_file_paths)
    mock_integrate_1d(map_controller)
    map_controller.load_btn_clicked()
    load_call_count = map_controller.model.calibration_model.integrate_1d.call_count

    # select second file in file list
    map_controller.widget.map_plot_widget.mouse_left_clicked.emit(2, 2)
    assert (
        map_controller.model.current_configuration.img_model.filename
        == map_img_file_paths[2]
    )
    assert map_controller.widget.control_widget.file_list.currentRow() == 2

    # check that integrate_1d was called only once
    assert (
        map_controller.model.calibration_model.integrate_1d.call_count
        == load_call_count + 1
    )


def test_click_in_pattern_will_update_region_of_interest(map_controller):
    click_pos = 30
    map_controller.widget.pattern_plot_widget.mouse_left_clicked.emit(click_pos, 10)
    assert (
        map_controller.widget.pattern_plot_widget.map_interactive_roi.center
        == approx(click_pos)
    )


def test_pattern_interactive_roi_updates_map(map_controller):
    map_controller.model.map_model.set_window = MagicMock()
    map_controller.widget.pattern_plot_widget.map_interactive_roi.sigRegionChanged.emit(
        SymmetricModifiedLinearRegionItem((10, 11))
    )
    map_controller.model.map_model.set_window.assert_called_once_with((10, 11))


def test_mouse_move_in_map_image_will_update_xyI(map_controller, map_model):
    mock_map_model(map_model)

    map_widget = map_controller.widget
    map_plot_widget = map_widget.map_plot_widget
    map_plot_control_widget = map_widget.map_plot_control_widget

    map_plot_widget.mouse_moved.emit(0, 1)
    assert map_plot_control_widget.mouse_x_label.text() == "X: 0"
    assert map_plot_control_widget.mouse_y_label.text() == "Y: 0"
    assert map_plot_control_widget.mouse_int_label.text() == "I: 1"

    # even when it is on fractional coordinates it should be rounded down
    map_plot_widget.mouse_moved.emit(0.7, 1.8)
    assert map_plot_control_widget.mouse_x_label.text() == "X: 0"
    assert map_plot_control_widget.mouse_y_label.text() == "Y: 0"
    assert map_plot_control_widget.mouse_int_label.text() == "I: 1"

    map_plot_widget.mouse_moved.emit(1, 0.7)
    assert map_plot_control_widget.mouse_x_label.text() == "X: 1"
    assert map_plot_control_widget.mouse_y_label.text() == "Y: 1"

    # it does not give coordinates, when outside of the map dimensions
    map_plot_widget.mouse_moved.emit(10, 10)
    assert map_plot_control_widget.mouse_x_label.text() == "X: "
    assert map_plot_control_widget.mouse_y_label.text() == "Y: "


def test_mouse_move_in_map_will_update_filename(map_controller, map_model):
    mock_map_model(map_model)
    map_controller.widget.map_plot_widget.mouse_moved.emit(0, 1)
    assert (
        map_controller.widget.map_plot_control_widget.filename_label.text()
        == map_img_file_names[0]
    )
    map_controller.widget.map_plot_widget.mouse_moved.emit(1, 1)
    assert (
        map_controller.widget.map_plot_control_widget.filename_label.text()
        == map_img_file_names[1]
    )
    map_controller.widget.map_plot_widget.mouse_moved.emit(0, 0)
    assert (
        map_controller.widget.map_plot_control_widget.filename_label.text()
        == map_img_file_names[3]
    )


def test_map_dimension_cb_updates_correctly(map_controller, map_model):
    map_model.window_intensities = np.array([1, 2, 3, 4, 5, 6])
    map_model.map = create_map(map_model.window_intensities, (2, 3))
    map_model.possible_dimensions = [(1, 6), (2, 3), (3, 2), (6, 1)]
    map_model.dimension = (2, 3)
    map_model.map_changed.emit()

    dim_cb = map_controller.widget.map_plot_control_widget.map_dimension_cb
    assert dim_cb.currentText() == "2x3"
    assert dim_cb.count() == 4
    assert dim_cb.currentIndex() == 1

    cb_items = [dim_cb.itemText(i) for i in range(dim_cb.count())]
    dimension_str = [f"{x}x{y}" for x, y in map_model.possible_dimensions]
    assert cb_items == dimension_str

    # changing dimensions:
    dim_cb.setCurrentIndex(2)
    assert map_model.dimension == (3, 2)
    assert map_model.map.shape == (3, 2)


def test_changing_configuration_updates_gui(map_controller, dioptas_model):
    load_calibration(map_controller)
    mock_open_filenames(map_img_file_paths)
    mock_integrate_1d(map_controller)
    map_controller.load_btn_clicked()

    map_config0 = map_controller.widget.map_plot_widget.img_data.copy()

    dioptas_model.add_configuration()
    assert dioptas_model.map_model.map is None

    load_calibration(map_controller)
    mock_open_filenames(list(reversed(map_img_file_paths)))
    map_controller.load_btn_clicked()

    assert dioptas_model.configurations[1].map_model.map is not None
    map_config1 = map_controller.widget.map_plot_widget.img_data.copy()
    assert np.array_equal(map_config1, map_config1)

    dioptas_model.select_configuration(0)
    assert map_controller.widget.map_plot_widget.img_data is not None
    assert np.array_equal(map_controller.widget.map_plot_widget.img_data, map_config0)

    items_text = [
        map_controller.widget.control_widget.file_list.item(i).text()
        for i in range(map_controller.widget.control_widget.file_list.count())
    ]
    assert items_text == map_img_file_names

    dioptas_model.select_configuration(1)
    assert np.array_equal(map_controller.widget.map_plot_widget.img_data, map_config1)

    items_text = [
        map_controller.widget.control_widget.file_list.item(i).text()
        for i in range(map_controller.widget.control_widget.file_list.count())
    ]
    assert items_text == list(reversed(map_img_file_names))


def test_progress_dialog_is_shown(map_controller):
    load_calibration(map_controller)
    mock_open_filenames(map_img_file_paths)
    mock_integrate_1d(map_controller)

    QtWidgets.QProgressDialog.setValue = MagicMock()
    map_controller.load_btn_clicked()

    # Progress dialog should be updated at least once (callbacks are throttled to
    # avoid GUI overhead, so the count may be less than the number of images)
    assert QtWidgets.QProgressDialog.setValue.call_count >= 1


def test_phase_is_displayed(map_controller, dioptas_model):
    load_calibration(map_controller)
    mock_integrate_1d(map_controller)
    mock_open_filenames(map_img_file_paths)
    map_controller.load_btn_clicked()

    pattern_widget = map_controller.widget.pattern_plot_widget

    assert pattern_widget.phases == []

    dioptas_model.phase_model.add_jcpds(os.path.join(data_path, "jcpds", "ar.jcpds"))
    assert len(pattern_widget.phases) == 1


def test_overlay_is_displayed(map_controller, dioptas_model: DioptasModel):
    dioptas_model.overlay_model.add_overlay(np.arange(10), np.arange(10), "test")
    pattern_widget = map_controller.widget.pattern_plot_widget

    assert len(pattern_widget.overlays) == 1


def test_green_line_in_pattern_plot(map_controller, dioptas_model):
    pattern_widget = map_controller.widget.pattern_plot_widget

    current_value = pattern_widget.get_pos_line()
    assert current_value is 0

    dioptas_model.clicked_tth_changed.emit(10)
    assert dioptas_model.clicked_tth == 10
    assert pattern_widget.get_pos_line() == 10

    # change unit, so that position of the line needs to
    # be in new unit
    dioptas_model.integration_unit = "q_A^-1"
    dioptas_model.clicked_tth_changed.emit(10)
    assert dioptas_model.clicked_tth == 10
    assert pattern_widget.get_pos_line() != 10


def test_green_line_shown_in_image(map_controller, dioptas_model):
    load_calibration(map_controller)
    mock_open_filenames(map_img_file_paths[:1])
    map_controller.load_btn_clicked()

    img_widget = map_controller.widget.img_plot_widget
    circle_plot_item = img_widget.circle_plot_items[0]
    x, y = circle_plot_item.getData()
    assert x is None
    assert y is None

    dioptas_model.clicked_tth_changed.emit(10)
    x, y = circle_plot_item.getData()
    assert len(x) > 0
    assert len(y) > 0


def test_green_line_shown_in_image_without_calibration(map_controller, dioptas_model):
    img_widget = map_controller.widget.img_plot_widget
    circle_plot_item = img_widget.circle_plot_items[0]

    dioptas_model.clicked_tth_changed.emit(10)
    x, y = circle_plot_item.getData()
    assert x is None
    assert y is None


def test_clicking_image_updates_tth_and_azi(map_controller, dioptas_model):
    load_calibration(map_controller)
    mock_open_filenames(map_img_file_paths[:1])
    map_controller.load_btn_clicked()

    img_widget = map_controller.widget.img_plot_widget
    assert dioptas_model.clicked_tth == 0
    assert dioptas_model.clicked_azi == 0
    img_widget.mouse_left_clicked.emit(100, 100)

    assert dioptas_model.clicked_tth != 0
    assert dioptas_model.clicked_azi != 0


def test_pattern_mouse_move_displays_positions(
    map_controller: MapController, dioptas_model: DioptasModel
):
    pattern_widget = map_controller.widget.pattern_plot_widget
    pos_widget = (
        map_controller.widget.pattern_footer_widget.mouse_unit_widget.cur_unit_widget
    )
    assert pos_widget.tth_lbl.text() == "2θ:"
    pattern_widget.mouse_moved.emit(10, 20)
    assert pos_widget.tth_lbl.text() == "2θ:%9.3f" % 10


def test_img_mouse_move_displays_positions(
    map_controller: MapController, dioptas_model: DioptasModel
):
    img_widget = map_controller.widget.img_plot_widget
    pos_widget = (
        map_controller.widget.pattern_footer_widget.mouse_unit_widget.cur_unit_widget
    )
    image_x = map_controller.widget.map_plot_control_widget.mouse_x_label
    image_y = map_controller.widget.map_plot_control_widget.mouse_y_label
    image_int = map_controller.widget.map_plot_control_widget.mouse_int_label
    assert pos_widget.tth_lbl.text() == "2θ:"

    img_widget.mouse_moved.emit(10, 20)
    assert image_x.text() == "X: 10"
    assert image_y.text() == "Y: 20"
    assert image_int.text() == "I: 0"

    load_calibration(map_controller)
    dioptas_model.img_model.load(map_img_file_paths[0])
    img_widget.mouse_moved.emit(100, 200)
    tth = dioptas_model.calibration_model.get_two_theta_img(200, 100)
    azi = dioptas_model.calibration_model.get_azi_img(200, 100)
    assert pos_widget.tth_lbl.text() == "2θ:%9.3f" % np.rad2deg(tth)
    assert pos_widget.azi_lbl.text() == "X:%9.3f" % np.rad2deg(azi)


def test_change_integration_unit(
    map_controller: MapController, dioptas_model: DioptasModel
):
    pattern_widget = map_controller.widget.pattern_plot_widget
    pattern_plot = pattern_widget.pattern_plot
    mock_integrate_1d(map_controller)
    load_calibration(map_controller)

    assert dioptas_model.integration_unit == "2th_deg"
    assert pattern_plot.getAxis("bottom").labelText == "2θ"
    assert pattern_plot.getAxis("bottom").labelUnits == "°"

    dioptas_model.integration_unit = "q_A^-1"
    assert pattern_plot.getAxis("bottom").labelText == "Q"
    assert pattern_plot.getAxis("bottom").labelUnits == "Å⁻¹"

    dioptas_model.integration_unit = "d_A"
    assert pattern_plot.getAxis("bottom").labelText == "d"
    assert pattern_plot.getAxis("bottom").labelUnits == "Å"

    dioptas_model.integration_unit = "2th_deg"
    assert pattern_plot.getAxis("bottom").labelText == "2θ"
    assert pattern_plot.getAxis("bottom").labelUnits == "°"


@pytest.mark.parametrize("file_type", ["png", "tiff", "txt"])
def test_save_map(map_controller, dioptas_model, tmp_path, file_type):
    mock_map_model(dioptas_model.map_model)
    map_widget = map_controller.widget

    filename = tmp_path / f"test_map.{file_type}"
    mock_save_filename(filename)

    map_widget.map_plot_control_widget.save_map_btn.clicked.emit()
    assert filename.exists()



from unittest.mock import MagicMock, patch
import tempfile
import pytest
import os

import h5py

from dioptas.model.Configuration import Configuration
from dioptas.model.MapModel import MapModel
from dioptas.model.util.integration import iter_frames_sequential
from dioptas.tests.utility import unittest_data_path
import numpy as np

jcpds_path = os.path.join(unittest_data_path, "jcpds")
map_img_path = os.path.join(unittest_data_path, "map")
map_pattern_path = os.path.join(unittest_data_path, "map", "xy")
map_img_file_names = [
    f for f in os.listdir(map_img_path) if os.path.isfile(os.path.join(map_img_path, f))
]
map_img_file_paths = [
    os.path.join(map_img_path, filename) for filename in map_img_file_names
]

multi_file_img_path = os.path.join(
    unittest_data_path, "lambda", "testasapo1_1009_00002_m1_part00000.nxs"
)


@pytest.fixture
def configuration() -> Configuration:
    return Configuration()


@pytest.fixture
def map_model(configuration: Configuration) -> MapModel:
    return MapModel(configuration)


def test_create_map(map_model: MapModel, configuration: Configuration):
    configuration.calibration_model.load(
        os.path.join(unittest_data_path, "CeO2_Pilatus1M.poni")
    )
    map_model.load(map_img_file_paths)
    assert map_model.filepaths == map_img_file_paths
    assert len(map_model.pattern_intensities) == len(map_img_file_paths)
    assert map_model.dimension == (3, 3)
    assert map_model.map.shape == (3, 3)


def test_load_empty_filelist(map_model: MapModel, configuration: Configuration):
    with pytest.raises(ValueError):
        map_model.load([])


def test_load_files_with_different_dimensions(
    map_model: MapModel, configuration: Configuration
):
    file_paths = [
        os.path.join(unittest_data_path, "CeO2_Pilatus1M.tif"),
        os.path.join(unittest_data_path, "image_001.tif"),
    ]
    configuration.calibration_model.load(
        os.path.join(unittest_data_path, "CeO2_Pilatus1M.poni")
    )
    with pytest.raises(ValueError):
        map_model.load(file_paths)
    
    assert map_model.filepaths is None
    assert map_model.pattern_intensities is None
    assert map_model.dimension is None
    assert configuration.img_model.img_changed.blocked is False
    assert configuration.trim_trailing_zeros is True


def test_set_dimensions(map_model: MapModel, configuration: Configuration):
    configuration.calibration_model.load(
        os.path.join(unittest_data_path, "CeO2_Pilatus1M.poni")
    )
    map_model.load(map_img_file_paths[:6])
    assert len(map_model.pattern_intensities) == 6
    assert map_model.dimension == (2, 3)
    assert map_model.map.shape == (2, 3)

    map_model.set_dimension((3, 2))
    assert map_model.dimension == (3, 2)
    assert map_model.map.shape == (3, 2)

    map_model.set_dimension((1, 6))
    assert map_model.dimension == (1, 6)
    assert map_model.map.shape == (1, 6)


def test_set_wrong_dimensions(map_model: MapModel, configuration: Configuration):
    configuration.calibration_model.load(
        os.path.join(unittest_data_path, "CeO2_Pilatus1M.poni")
    )
    map_model.load(map_img_file_paths[:6])
    assert len(map_model.pattern_intensities) == 6
    assert map_model.dimension == (2, 3)
    assert map_model.map.shape == (2, 3)

    map_model.set_dimension((3, 3))
    assert map_model.dimension == (2, 3)
    assert map_model.map.shape == (2, 3)


def test_set_different_window(map_model: MapModel, configuration: Configuration):
    configuration.calibration_model.load(
        os.path.join(unittest_data_path, "CeO2_Pilatus1M.poni")
    )
    map_model.load(map_img_file_paths[:6])

    map_model.set_window((15, 16))
    assert map_model.window_intensities.all() > 0

    map_model.set_window((35, 40))  # window outside of pattern range
    assert map_model.window_intensities.all() == 0


def test_get_point_information(map_model: MapModel, configuration: Configuration):
    configuration.calibration_model.load(
        os.path.join(unittest_data_path, "CeO2_Pilatus1M.poni")
    )
    map_model.load(map_img_file_paths[:6])

    assert map_model.dimension == (2, 3)
    for i in range(6):
        column_index = i % 3
        row_index = i // 3
        point_info = map_model.get_point_info(row_index, column_index)
        assert point_info.filename == map_img_file_names[i]


def test_use_multi_file_img(map_model: MapModel, configuration: Configuration):
    configuration.calibration_model.load(
        os.path.join(unittest_data_path, "CeO2_Pilatus1M.poni")
    )
    map_model.load([multi_file_img_path])

    img_model = configuration.img_model

    assert map_model.filepaths == [multi_file_img_path]
    assert img_model.series_max == 10
    assert map_model.pattern_intensities.shape[0] == 10


def test_integrates_each_image_only_once(
    map_model: MapModel, configuration: Configuration
):
    configuration.calibration_model.load(
        os.path.join(unittest_data_path, "CeO2_Pilatus1M.poni")
    )
    x = np.linspace(0, 10, 100)
    y = np.sin(x)
    configuration.calibration_model.integrate_1d = MagicMock(return_value=(x, y))
    configuration.calibration_model.can_use_dioptrin_batch = MagicMock(return_value=False)
    map_model.load(map_img_file_paths)

    assert configuration.calibration_model.integrate_1d.call_count == len(
        map_img_file_paths
    )


def test_emits_point_integrated_signal(
    map_model: MapModel, configuration: Configuration
):
    configuration.calibration_model.load(
        os.path.join(unittest_data_path, "CeO2_Pilatus1M.poni")
    )
    # mock the integrate_1d method
    x = np.linspace(0, 10, 100)
    y = np.sin(x)
    configuration.calibration_model.integrate_1d = MagicMock(return_value=(x, y))

    listener = MagicMock()
    map_model.point_integrated.connect(listener)

    map_model.load(map_img_file_paths)
    assert listener.call_count == 9
    # assert listener.call_args_list == [(1), (2), (3), (4), (5), (6), (7), (8), (9)]


def test_emits_point_integrated_signal_with_multiimage_file(
    map_model: MapModel, configuration: Configuration
):
    configuration.calibration_model.load(
        os.path.join(unittest_data_path, "CeO2_Pilatus1M.poni")
    )
    # mock the integrate_1d method
    x = np.linspace(0, 10, 100)
    y = np.sin(x)
    configuration.calibration_model.integrate_1d = MagicMock(return_value=(x, y))

    listener = MagicMock()
    map_model.point_integrated.connect(listener)

    map_model.load([multi_file_img_path])
    assert listener.call_count == 10


def test_iter_frames_sequential_bitshuffle_shape_mismatch(configuration: Configuration):
    """Bitshuffle HDF5 path in iter_frames_sequential must reject shape mismatches."""
    img_model = configuration.img_model

    expected_shape = (2048, 2048)
    wrong_shape = (3262, 3108)

    mock_loader = MagicMock()
    mock_loader.gen_frames.return_value = iter([np.zeros(wrong_shape)])

    with patch(
        "dioptas.model.util.integration.try_open_bitshuffle_hdf5",
        return_value=mock_loader,
    ):
        gen = iter_frames_sequential(
            img_model,
            ["/fake/file.h5"],
            img_shape=expected_shape,
        )
        with pytest.raises(ValueError, match="expected"):
            next(gen)


def test_save_load_hdf5_round_trip(map_model: MapModel, configuration: Configuration):
    """Map state survives a save/load round-trip through HDF5."""
    configuration.calibration_model.load(
        os.path.join(unittest_data_path, "CeO2_Pilatus1M.poni")
    )
    map_model.load(map_img_file_paths)

    # tweak window and dimension so we verify they are restored
    map_model.set_window((15, 20))
    map_model.set_dimension((3, 3))

    with tempfile.NamedTemporaryFile(suffix=".h5", delete=False) as tmp:
        tmp_path = tmp.name

    try:
        # save
        with h5py.File(tmp_path, "w") as f:
            map_model.save_in_hdf5(f)

        # load into a fresh model
        new_model = MapModel(configuration)
        with h5py.File(tmp_path, "r") as f:
            new_model.load_from_hdf5(f)

        np.testing.assert_array_equal(new_model.pattern_x, map_model.pattern_x)
        np.testing.assert_array_equal(
            new_model.pattern_intensities, map_model.pattern_intensities
        )
        assert new_model.filepaths == map_model.filepaths
        assert len(new_model.point_infos) == len(map_model.point_infos)
        for a, b in zip(new_model.point_infos, map_model.point_infos):
            assert a.filepath == b.filepath
            assert a.frame_index == b.frame_index
        assert new_model.window == pytest.approx(map_model.window)
        assert new_model.dimension == map_model.dimension
        np.testing.assert_array_equal(new_model.map, map_model.map)
    finally:
        os.unlink(tmp_path)


def test_load_hdf5_without_map_group(map_model: MapModel):
    """Loading from an HDF5 file with no 'map' group is a no-op."""
    with tempfile.NamedTemporaryFile(suffix=".h5", delete=False) as tmp:
        tmp_path = tmp.name

    try:
        with h5py.File(tmp_path, "w") as f:
            pass  # empty file

        map_model.load_from_hdf5(h5py.File(tmp_path, "r"))
        assert map_model.filepaths is None
        assert map_model.map is None
    finally:
        os.unlink(tmp_path)



# SPDX-License-Identifier: MIT

import os
import pytest
from unittest.mock import MagicMock

import numpy as np

from ..utility import click_button
from ...model.util.HelperModule import get_partial_value
from .test_BatchController_part1 import *

unittest_data_path = os.path.join(os.path.dirname(__file__), "../data")
jcpds_path = os.path.join(unittest_data_path, "jcpds")


def test_set_unit(
    batch_controller, batch_widget, integration_widget, dioptas_model, load_proc_data
):
    batch_widget.activate_stack_plot()
    bottom_axis = batch_widget.stack_plot_widget.img_view.bottom_axis_cake

    class DummyViewRect(object):
        _width = 3500
        _left = 5

        def width(self):
            return self._width

        def left(self):
            return self._left

    batch_widget.stack_plot_widget.img_view.img_view_rect = MagicMock(
        return_value=DummyViewRect()
    )

    batch_controller.set_unit_tth()
    assert dioptas_model.current_configuration.integration_unit == "2th_deg"
    assert bottom_axis._tickLevels is None
    assert bottom_axis.range[0] == pytest.approx(9.7129, 0.001)

    click_button(batch_widget.options_widget.q_btn)
    assert bottom_axis._tickLevels is not None
    assert bottom_axis._tickLevels[0][0][0] == pytest.approx(22.092006, 0.001)

    batch_controller.set_unit_tth()
    assert dioptas_model.current_configuration.integration_unit == "2th_deg"
    assert bottom_axis._tickLevels is None

    batch_controller.set_unit_d()
    assert integration_widget.integration_pattern_widget.d_btn.isChecked()
    assert dioptas_model.current_configuration.integration_unit == "d_A"
    assert bottom_axis._tickLevels[0][0][0] == pytest.approx(10.484, 0.001)


def test_show_phases(
    batch_controller, phase_controller, batch_widget, dioptas_model, load_proc_data
):
    # Load phases
    dioptas_model.phase_model.add_jcpds(os.path.join(jcpds_path, "FeGeO3_cpx.jcpds"))
    assert str(batch_widget.control_widget.phases_btn.text()) == "Show Phases"
    batch_widget.control_widget.phases_btn.setChecked(True)
    batch_controller.toggle_show_phases()
    assert str(batch_widget.control_widget.phases_btn.text()) == "Hide Phases"

    assert len(batch_widget.stack_plot_widget.img_view.phases) == 1
    assert len(batch_widget.stack_plot_widget.img_view.phases[0].line_items) == 27


def test_load_single_image(batch_controller, batch_widget, load_proc_data):
    batch_widget.activate_stack_plot()
    batch_controller.load_single_image(10, 15)

    assert (
        batch_widget.position_widget.mouse_pos_widget.clicked_pos_widget.x_pos_lbl.text()
        == "Img: 15"
    )

    filename = os.path.join(
        unittest_data_path, "lambda", "testasapo1_1009_00002_m1_part00001.nxs"
    )
    assert batch_widget.windowTitle() == f"Batch widget. {filename} - 5"
    assert batch_widget.stack_plot_widget.img_view.horizontal_line.value() == 15


def test_plot_image(batch_controller, batch_widget, dioptas_model, load_proc_data):
    batch_widget.activate_stack_plot()
    batch_controller.plot_image(15)

    filename = os.path.join(
        unittest_data_path, "lambda", "testasapo1_1009_00002_m1_part00001.nxs"
    )
    assert batch_widget.windowTitle() == f"Batch widget. {filename} - 5"
    assert dioptas_model.current_configuration.auto_integrate_pattern
    assert (
        batch_widget.position_widget.mouse_pos_widget.clicked_pos_widget.x_pos_lbl.text()
        == f"Img: 15"
    )


def test_plot_pattern(batch_controller, dioptas_model, load_proc_data):
    batch_controller.plot_pattern(10, 15)

    assert dioptas_model.pattern_model.pattern.data[0][0] == pytest.approx(
        9.6926780, 0.001
    )
    assert dioptas_model.pattern_model.pattern.data[1][0] == np.float32(0.1)


def test_update_y_axis(batch_controller, batch_widget, load_proc_data):
    batch_widget.activate_stack_plot()
    batch_widget.position_widget.step_series_widget.slider.setValue(15)
    batch_widget.position_widget.step_series_widget.start_txt.blockSignals(True)
    batch_widget.position_widget.step_series_widget.stop_txt.blockSignals(True)
    batch_widget.position_widget.step_series_widget.start_txt.setValue(5)
    batch_widget.position_widget.step_series_widget.stop_txt.setValue(28)
    batch_widget.position_widget.step_series_widget.start_txt.blockSignals(False)
    batch_widget.position_widget.step_series_widget.stop_txt.blockSignals(False)
    batch_controller.plot_batch()

    batch_controller.update_y_axis()
    assert batch_widget.stack_plot_widget.img_view.left_axis_cake.range[
        0
    ] == pytest.approx(3.511, abs=0.1)
    assert batch_widget.stack_plot_widget.img_view.left_axis_cake.range[
        1
    ] == pytest.approx(29.678, abs=0.1)


def test_click_in_2d_widget_sends_clicked_change(
    batch_controller, batch_widget, dioptas_model, load_proc_data
):
    dioptas_model.clicked_tth_changed.emit = MagicMock()
    batch_widget.stack_plot_widget.img_view.mouse_left_clicked.emit(20, 10)
    dioptas_model.clicked_tth_changed.emit.assert_called_once_with(
        get_partial_value(dioptas_model.batch_model.binning, 20 - 0.5)
    )


def test_click_in_2d_widget_out_of_range_does_not_send_clicked_change(
    batch_controller, batch_widget, dioptas_model, load_proc_data
):
    dioptas_model.clicked_tth_changed.emit = MagicMock()
    batch_widget.stack_plot_widget.img_view.mouse_left_clicked.emit(-1, 10)
    dioptas_model.clicked_tth_changed.emit.assert_not_called()


def test_clicked_tth_change_signal_changes_line_pos(
    batch_controller, batch_widget, dioptas_model, load_proc_data
):
    batch_controller.plot_batch()

    # in plot range
    dioptas_model.clicked_tth_changed.emit(12)
    new_line_pos = batch_widget.stack_plot_widget.img_view.vertical_line.pos()[0]
    assert get_partial_value(
        dioptas_model.batch_model.binning, new_line_pos
    ) == pytest.approx(12, 0.01)

    #  lower than range
    dioptas_model.clicked_tth_changed.emit(8)
    assert (
        batch_widget.stack_plot_widget.img_view.vertical_line
        not in batch_widget.stack_plot_widget.img_view.img_view_box.addedItems
    )

    # in range again
    dioptas_model.clicked_tth_changed.emit(12)
    assert (
        batch_widget.stack_plot_widget.img_view.vertical_line
        in batch_widget.stack_plot_widget.img_view.img_view_box.addedItems
    )
    new_line_pos = batch_widget.stack_plot_widget.img_view.vertical_line.pos()[0]
    assert get_partial_value(
        dioptas_model.batch_model.binning, new_line_pos
    ) == pytest.approx(12, 0.01)

    # larger than range
    dioptas_model.clicked_tth_changed.emit(35)
    assert (
        batch_widget.stack_plot_widget.img_view.vertical_line
        not in batch_widget.stack_plot_widget.img_view.img_view_box.addedItems
    )


def test_set_navigation_range(batch_controller, batch_widget, load_proc_data):
    batch_controller.set_navigation_range((10, 50))
    assert batch_widget.position_widget.step_series_widget.start_txt.text() == "10"
    assert batch_widget.position_widget.step_series_widget.stop_txt.text() == "50"
    assert batch_widget.position_widget.step_series_widget.slider.value() == 10

    batch_widget.position_widget.step_series_widget.slider.setValue(50)
    batch_controller.set_navigation_range((10, 50))
    assert batch_widget.position_widget.step_series_widget.start_txt.text() == "10"
    assert batch_widget.position_widget.step_series_widget.stop_txt.text() == "50"
    assert batch_widget.position_widget.step_series_widget.slider.value() == 50

    # ToDo Test interaction with other controllers: Integration window vertical line. Patterns


@pytest.mark.skip("needs to be fixed")
def test_integrate(
    batch_controller, batch_widget, integration_widget, dioptas_model, load_proc_data
):
    batch_widget.position_widget.step_series_widget.start_txt.blockSignals(True)
    batch_widget.position_widget.step_series_widget.stop_txt.blockSignals(True)
    batch_widget.position_widget.step_series_widget.start_txt.setValue(5)
    batch_widget.position_widget.step_series_widget.stop_txt.setValue(28)
    batch_widget.position_widget.step_series_widget.step_txt.setValue(3)
    batch_widget.position_widget.step_series_widget.start_txt.blockSignals(False)
    batch_widget.position_widget.step_series_widget.stop_txt.blockSignals(False)

    batch_widget.mode_widget.view_f_btn.setChecked(False)
    integration_widget.automatic_binning_cb.setChecked(False)
    integration_widget.bin_count_txt.setText(str(4000))

    dioptas_model.calibration_model.load(
        os.path.join(unittest_data_path, "lambda", "L2.poni")
    )

    batch_controller.integrate()

    assert dioptas_model.batch_model.data.shape == (8, 3984)
    assert dioptas_model.batch_model.binning.shape == (3984,)
    assert dioptas_model.batch_model.n_img == 8
    assert dioptas_model.batch_model.n_img_all == 50
    assert dioptas_model.batch_model.pos_map.shape == (8, 2)



# SPDX-License-Identifier: MIT

from .BackgroundController import BackgroundController
from .ImageController import ImageController
from .IntegrationController import IntegrationController
from .phase.JcpdsEditorController import JcpdsEditorController
from .overlay.OverlayController import OverlayController
from .phase.PhaseController import PhaseController
from .phase.PhaseController import PhaseInPatternController
from .phase.PhaseInCakeController import PhaseInCakeController
from .PatternController import PatternController
from .OptionsController import OptionsController
from .BatchController import BatchController
